# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import json
import os
import sys
from typing import cast
from unittest import mock

import ddt

from osprofiler.cmd import shell
from osprofiler import exc
from osprofiler.tests import test


@ddt.ddt
class ShellTestCase(test.TestCase):
    TRACE_ID = "c598094d-bbee-40b6-b317-d76003b679d3"

    def setUp(self):
        super().setUp()
        self.old_environment = os.environ.copy()

    def tearDown(self):
        super().tearDown()
        os.environ.clear()
        os.environ.update(self.old_environment)

    def _trace_show_cmd(self, format_=None):
        cmd = f"trace show --connection-string redis:// {self.TRACE_ID}"
        return cmd if format_ is None else f"{cmd} --{format_}"

    @mock.patch("sys.stdout", io.StringIO())
    @mock.patch("osprofiler.cmd.shell.OSProfilerShell")
    def test_shell_main(self, mock_shell):
        mock_shell.side_effect = exc.CommandError("some_message")
        shell.main()
        self.assertEqual(
            "some_message\n", cast(io.StringIO, sys.stdout).getvalue()
        )

    def run_command(self, cmd):
        shell.OSProfilerShell(cmd.split())

    def _test_with_command_error(self, cmd, expected_message):
        try:
            self.run_command(cmd)
        except exc.CommandError as actual_error:
            self.assertEqual(str(actual_error), expected_message)
        else:
            raise ValueError(
                "Expected: `osprofiler.exc.CommandError` is raised with "
                f"message: '{expected_message}'."
            )

    @mock.patch("osprofiler.drivers.redis_driver.Redis.get_report")
    def test_trace_show_no_selected_format(self, mock_get):
        mock_get.return_value = self._create_mock_notifications()
        msg = (
            "You should choose one of the following output formats: "
            "json, html or dot."
        )
        self._test_with_command_error(self._trace_show_cmd(), msg)

    @mock.patch("osprofiler.drivers.redis_driver.Redis.get_report")
    @ddt.data(
        None,
        {
            "info": {"started": 0, "finished": 1, "name": "total"},
            "children": [],
        },
    )
    def test_trace_show_trace_id_not_found(self, notifications, mock_get):
        mock_get.return_value = notifications

        msg = (
            f"Trace with UUID {self.TRACE_ID} not found. Please check the "
            f"HMAC key used in the command."
        )

        self._test_with_command_error(self._trace_show_cmd(), msg)

    def _create_mock_notifications(self):
        notifications = {
            "info": {"started": 0, "finished": 1, "name": "total"},
            "children": [
                {
                    "info": {"started": 0, "finished": 1, "name": "total"},
                    "children": [],
                }
            ],
        }
        return notifications

    @mock.patch("sys.stdout", io.StringIO())
    @mock.patch("osprofiler.drivers.redis_driver.Redis.get_report")
    def test_trace_show_in_json(self, mock_get):
        notifications = self._create_mock_notifications()
        mock_get.return_value = notifications

        self.run_command(self._trace_show_cmd(format_="json"))
        self.assertEqual(
            "{}\n".format(
                json.dumps(
                    notifications,
                    indent=2,
                    separators=(",", ": "),
                )
            ),
            cast(io.StringIO, sys.stdout).getvalue(),
        )

    @mock.patch("sys.stdout", io.StringIO())
    @mock.patch("osprofiler.drivers.redis_driver.Redis.get_report")
    def test_trace_show_in_html(self, mock_get):
        notifications = self._create_mock_notifications()
        mock_get.return_value = notifications

        # NOTE(akurilin): to simplify assert statement, html-template should be
        # replaced.
        html_template = (
            "A long time ago in a galaxy far, far away..."
            "    some_data = $DATA"
            "It is a period of civil war. Rebel"
            "spaceships, striking from a hidden"
            "base, have won their first victory"
            "against the evil Galactic Empire."
        )

        with mock.patch(
            "osprofiler.cmd.commands.open",
            mock.mock_open(read_data=html_template),
            create=True,
        ):
            self.run_command(self._trace_show_cmd(format_="html"))
            self.assertEqual(
                "A long time ago in a galaxy far, far away..."
                "    some_data = {}"
                "It is a period of civil war. Rebel"
                "spaceships, striking from a hidden"
                "base, have won their first victory"
                "against the evil Galactic Empire."
                "\n".format(
                    json.dumps(notifications, indent=4, separators=(",", ": "))
                ),
                cast(io.StringIO, sys.stdout).getvalue(),
            )

    @mock.patch("sys.stdout", io.StringIO())
    @mock.patch("osprofiler.drivers.redis_driver.Redis.get_report")
    def test_trace_show_write_to_file(self, mock_get):
        notifications = self._create_mock_notifications()
        mock_get.return_value = notifications

        with mock.patch(
            "osprofiler.cmd.commands.open", mock.mock_open(), create=True
        ) as mock_open:
            self.run_command(
                "{} --out='/file'".format(self._trace_show_cmd(format_="json"))
            )

            output = mock_open.return_value.__enter__.return_value
            output.write.assert_called_once_with(
                json.dumps(notifications, indent=2, separators=(",", ": "))
            )



# Copyright (c) 2016 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os

from oslo_config import cfg

from osprofiler.drivers import base
from osprofiler import initializer
from osprofiler import opts
from osprofiler import profiler
from osprofiler.tests import test


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


@profiler.trace_cls("rpc", hide_args=True)
class Foo:
    def bar(self, x):
        return self.baz(x, x)

    def baz(self, x, y):
        return x * y


class DriverTestCase(test.FunctionalTestCase):
    SERVICE = "service"
    PROJECT = "project"

    def setUp(self):
        super().setUp()
        CONF(["--config-file", os.path.dirname(__file__) + "/config.cfg"])
        opts.set_defaults(
            CONF, enabled=True, trace_sqlalchemy=False, hmac_keys="SECRET_KEY"
        )

    def _assert_dict(self, info, **kwargs):
        for key in kwargs:
            self.assertEqual(kwargs[key], info[key])

    def _assert_child_dict(self, child, base_id, parent_id, name, fn_name):
        self.assertEqual(parent_id, child["parent_id"])

        exp_info = {
            "name": "rpc",
            "service": self.SERVICE,
            "project": self.PROJECT,
        }
        self._assert_dict(child["info"], **exp_info)

        raw_start = child["info"][f"meta.raw_payload.{name}-start"]
        self.assertEqual(fn_name, raw_start["info"]["function"]["name"])
        exp_raw = {
            "name": f"{name}-start",
            "service": self.SERVICE,
            "trace_id": child["trace_id"],
            "project": self.PROJECT,
            "base_id": base_id,
        }
        self._assert_dict(raw_start, **exp_raw)

        raw_stop = child["info"][f"meta.raw_payload.{name}-stop"]
        exp_raw["name"] = f"{name}-stop"
        self._assert_dict(raw_stop, **exp_raw)

    def test_get_report(self):
        # initialize profiler notifier (the same way as in services)
        initializer.init_from_conf(
            CONF, {}, self.PROJECT, self.SERVICE, "host"
        )
        profiler.init("SECRET_KEY")

        # grab base_id
        p = profiler.get()
        assert p is not None  # noqa: S101
        base_id = p.get_base_id()

        # execute profiled code
        foo = Foo()
        foo.bar(1)

        # instantiate report engine (the same way as in osprofiler CLI)
        engine = base.get_driver(
            CONF.profiler.connection_string,
            project=self.PROJECT,
            service=self.SERVICE,
            host="host",
            conf=CONF,
        )

        # generate the report
        report = engine.get_report(base_id)
        LOG.debug("OSProfiler report: %s", report)

        # verify the report
        self.assertEqual("total", report["info"]["name"])
        self.assertEqual(2, report["stats"]["rpc"]["count"])
        self.assertEqual(1, len(report["children"]))

        cbar = report["children"][0]
        self._assert_child_dict(
            cbar,
            base_id,
            base_id,
            "rpc",
            "osprofiler.tests.functional.test_driver.Foo.bar",
        )

        self.assertEqual(1, len(cbar["children"]))
        cbaz = cbar["children"][0]
        self._assert_child_dict(
            cbaz,
            base_id,
            cbar["trace_id"],
            "rpc",
            "osprofiler.tests.functional.test_driver.Foo.baz",
        )


class RedisDriverTestCase(DriverTestCase):
    def setUp(self):
        super(DriverTestCase, self).setUp()
        CONF([])
        opts.set_defaults(
            CONF,
            connection_string="redis://localhost:6379",
            enabled=True,
            trace_sqlalchemy=False,
            hmac_keys="SECRET_KEY",
        )

    def test_list_traces(self):
        # initialize profiler notifier (the same way as in services)
        initializer.init_from_conf(
            CONF, {}, self.PROJECT, self.SERVICE, "host"
        )
        profiler.init("SECRET_KEY")

        # grab base_id
        p = profiler.get()
        assert p is not None  # noqa: S101
        base_id = p.get_base_id()

        # execute profiled code
        foo = Foo()
        foo.bar(1)

        # instantiate report engine (the same way as in osprofiler CLI)
        engine = base.get_driver(
            CONF.profiler.connection_string,
            project=self.PROJECT,
            service=self.SERVICE,
            host="host",
            conf=CONF,
        )

        # generate the report
        traces = engine.list_traces()
        LOG.debug("Collected traces: %s", traces)

        # ensure trace with base_id is in the list of traces
        self.assertIn(base_id, [t["base_id"] for t in traces])



# Copyright 2018 Fujitsu Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import Any

from oslo_config import cfg

from osprofiler.drivers import base
from osprofiler import exc


# TODO(tkajinam): Remove this and the deprecated options after G-release
class Jaeger(base.Driver):
    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        conf: cfg.ConfigOpts | None = None,
        **kwargs: Any,
    ) -> None:
        """Jaeger driver for OSProfiler."""

        raise exc.CommandError('Jaeger driver is no longer supported')

    @classmethod
    def get_name(cls) -> str:
        return "jaeger"



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Callable
import logging
from typing import Any

from osprofiler.drivers import base


LOG = logging.getLogger(__name__)


def _noop_notifier(info: dict[str, Any], context: Any = None) -> None:
    """Do nothing on notify()."""


# NOTE(boris-42): By default we are using noop notifier.
__notifier: Callable[..., None] = _noop_notifier
__notifier_cache: dict[
    str, Callable[..., None]
] = {}  # map: connection-string -> notifier


def notify(info: dict[str, Any]) -> None:
    """Passes the profiling info to the notifier callable.

    :param info: dictionary with profiling information
    """
    __notifier(info)


def get() -> Callable[..., None]:
    """Returns notifier callable."""
    return __notifier


def set(notifier: Callable[..., None]) -> None:
    """Service that are going to use profiler should set callable notifier.

    Callable notifier is instance of callable object, that accept exactly
    one argument "info". "info" - is dictionary of values that contains
    profiling information.
    """
    global __notifier
    __notifier = notifier


def create(
    connection_string: str, *args: Any, **kwargs: Any
) -> Callable[..., None]:
    """Create notifier based on specified plugin_name

    :param connection_string: connection string which specifies the storage
                              driver for notifier
    :param args: args that will be passed to the driver's __init__ method
    :param kwargs: kwargs that will be passed to the driver's __init__ method
    :returns: Callable notifier method
    """
    global __notifier_cache
    if connection_string not in __notifier_cache:
        try:
            driver = base.get_driver(connection_string, *args, **kwargs)
            __notifier_cache[connection_string] = driver.notify
            LOG.info(
                "osprofiler is enabled with connection string: %s",
                connection_string,
            )
        except Exception:
            LOG.exception(
                "Could not initialize driver for connection string "
                "%s, osprofiler is disabled",
                connection_string,
            )
            __notifier_cache[connection_string] = _noop_notifier

    return __notifier_cache[connection_string]


def clear_notifier_cache() -> None:
    __notifier_cache.clear()



# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from osprofiler.drivers import base
from osprofiler.tests import test


class MessagingTestCase(test.TestCase):
    @mock.patch("oslo_utils.importutils.try_import")
    def test_init_no_oslo_messaging(self, try_import_mock):
        try_import_mock.return_value = None

        self.assertRaises(
            ValueError,
            base.get_driver,
            "messaging://",
            project="project",
            service="service",
            host="host",
            context={},
        )

    @mock.patch("oslo_utils.importutils.try_import")
    def test_init_and_notify(self, try_import_mock):
        context = "context"
        transport = "transport"
        project = "project"
        service = "service"
        host = "host"

        # emulate dynamic load of oslo.messaging library
        oslo_messaging_mock = mock.Mock()
        try_import_mock.return_value = oslo_messaging_mock

        # mock oslo.messaging APIs
        notifier_mock = mock.Mock()
        oslo_messaging_mock.Notifier.return_value = notifier_mock
        oslo_messaging_mock.get_notification_transport.return_value = transport

        notify_func = base.get_driver(
            "messaging://",
            project=project,
            service=service,
            context=context,
            host=host,
        ).notify

        oslo_messaging_mock.Notifier.assert_called_once_with(
            transport,
            publisher_id=host,
            driver="messaging",
            topics=["profiler"],
            retry=0,
        )

        info = {"a": 10, "project": project, "service": service, "host": host}
        notify_func(info)

        notifier_mock.info.assert_called_once_with(
            context, "profiler.service", info
        )

        notifier_mock.reset_mock()
        notify_func(info, context="my_context")
        notifier_mock.info.assert_called_once_with(
            "my_context", "profiler.service", info
        )



# Copyright 2019 SUSE Linux GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from typing import Any

from oslo_serialization import jsonutils

from osprofiler.drivers import base
from osprofiler import exc

LOG = logging.getLogger(__name__)


class SQLAlchemyDriver(base.Driver):
    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            connection_str, project=project, service=service, host=host
        )

        try:
            from sqlalchemy import create_engine  # type: ignore[import-not-found]
            from sqlalchemy import Table, MetaData, Column
            from sqlalchemy import String, JSON, Integer
        except ImportError:
            LOG.exception("To use this command, install 'SQLAlchemy'")
        else:
            self._metadata = MetaData()
            self._data_table = Table(
                "data",
                self._metadata,
                Column("id", Integer, primary_key=True),
                # timestamp - date/time of the trace point
                Column("timestamp", String(26), index=True),
                # base_id - uuid common for all notifications related to
                # one trace
                Column("base_id", String(255), index=True),
                # parent_id - uuid of parent element in trace
                Column("parent_id", String(255), index=True),
                # trace_id - uuid of current element in trace
                Column("trace_id", String(255), index=True),
                Column("project", String(255), index=True),
                Column("host", String(255), index=True),
                Column("service", String(255), index=True),
                # name - trace point name
                Column("name", String(255), index=True),
                Column("data", JSON),
            )

        # we don't want to kill any service that does use osprofiler
        try:
            self._engine = create_engine(connection_str)
            self._conn = self._engine.connect()

            # FIXME(toabctl): Not the best idea to create the table on every
            # startup when using the sqlalchemy driver...
            self._metadata.create_all(self._engine, checkfirst=True)
        except Exception:
            LOG.exception(
                "Failed to create engine/connection and setup "
                "intial database tables"
            )

    @classmethod
    def get_name(cls) -> str:
        return "sqlalchemy"

    def notify(
        self, info: dict[str, Any], context: Any = None, **kwargs: Any
    ) -> None:
        """Write a notification the the database"""
        data = info.copy()
        base_id = data.pop("base_id", None)
        timestamp = data.pop("timestamp", None)
        parent_id = data.pop("parent_id", None)
        trace_id = data.pop("trace_id", None)
        project = data.pop("project", self.project)
        host = data.pop("host", self.host)
        service = data.pop("service", self.service)
        name = data.pop("name", None)

        try:
            ins = self._data_table.insert().values(
                timestamp=timestamp,
                base_id=base_id,
                parent_id=parent_id,
                trace_id=trace_id,
                project=project,
                service=service,
                host=host,
                name=name,
                data=jsonutils.dumps(data),
            )
            self._conn.execute(ins)
        except Exception:
            LOG.exception(
                "Can not store osprofiler tracepoint %s (base id %s)",
                trace_id,
                base_id,
            )

    def list_traces(
        self, fields: set[str] | None = None
    ) -> list[dict[str, Any]]:
        try:
            from sqlalchemy.sql import select  # type: ignore[import-not-found]
        except ImportError:
            raise exc.CommandError(
                "To use this command, you should install 'SQLAlchemy'"
            )
        fields = set(fields or self.default_trace_fields)
        stmt = select([self._data_table])
        seen_ids: set[str] = set()
        result: list[dict[str, Any]] = []
        traces = self._conn.execute(stmt).fetchall()
        for trace in traces:
            if trace["base_id"] not in seen_ids:
                seen_ids.add(trace["base_id"])
                result.append(
                    {
                        key: value
                        for key, value in trace.items()
                        if key in fields
                    }
                )
        return result

    def get_report(self, base_id: str) -> dict[str, Any]:
        try:
            from sqlalchemy.sql import select
        except ImportError:
            raise exc.CommandError(
                "To use this command, you should install 'SQLAlchemy'"
            )
        stmt = select([self._data_table]).where(
            self._data_table.c.base_id == base_id
        )
        results = self._conn.execute(stmt).fetchall()
        for n in results:
            timestamp = n["timestamp"]
            trace_id = n["trace_id"]
            parent_id = n["parent_id"]
            name = n["name"]
            project = n["project"]
            service = n["service"]
            host = n["host"]
            data = jsonutils.loads(n["data"])
            self._append_results(
                trace_id,
                parent_id,
                name,
                project,
                service,
                host,
                timestamp,
                data,
            )
        return self._parse_results()



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import hashlib
import hmac
import json
import os
import uuid
from collections.abc import Generator, Sequence
from typing import Any, TypeVar, overload

from oslo_utils import uuidutils

_C = TypeVar("_C")
_T = TypeVar("_T")

_AlreadySplit = TypeVar('_AlreadySplit', bound=list[Any] | tuple[Any, ...])


@overload
def split(text: str, strip: bool = True) -> list[str]: ...


@overload
def split(text: _AlreadySplit, strip: bool = True) -> _AlreadySplit: ...


def split(
    text: str | _AlreadySplit, strip: bool = True
) -> list[str] | _AlreadySplit:
    """Splits a comma separated text blob into its components.

    Does nothing if already a list or tuple.
    """
    if isinstance(text, (tuple, list)):
        return text
    if not isinstance(text, str):
        raise TypeError(f"Unknown how to split '{text}': {type(text)}")
    if strip:
        return [t.strip() for t in text.split(",") if t.strip()]
    else:
        return text.split(",")


def binary_encode(text: bytes | str, encoding: str = "utf-8") -> bytes:
    """Converts a string of into a binary type using given encoding.

    Does nothing if text not unicode string.
    """
    if isinstance(text, bytes):
        return text
    elif isinstance(text, str):
        return text.encode(encoding)
    else:
        raise TypeError("Expected binary or string type")


def binary_decode(data: bytes | str, encoding: str = "utf-8") -> str:
    """Converts a binary type into a text type using given encoding.

    Does nothing if data is already unicode string.
    """
    if isinstance(data, bytes):
        return data.decode(encoding)
    elif isinstance(data, str):
        return data
    else:
        raise TypeError("Expected binary or string type")


def generate_hmac(data: bytes | str, hmac_key: bytes | str) -> str:
    """Generate a hmac using a known key given the provided content."""
    h = hmac.new(binary_encode(hmac_key), digestmod=hashlib.sha1)
    h.update(binary_encode(data))
    return h.hexdigest()


def signed_pack(
    data: dict[str, str], hmac_key: str | None
) -> tuple[bytes, str | None]:
    """Pack and sign data with hmac_key."""
    raw_data = base64.urlsafe_b64encode(binary_encode(json.dumps(data)))

    # NOTE(boris-42): Don't generate_hmac if there is no hmac_key, mostly
    #                 security reason, we shouldn't allow to use WsgiMiddleware
    #                 without hmac_key, cause everybody will be able to trigger
    #                 profiler and organize DDOS.
    return raw_data, generate_hmac(raw_data, hmac_key) if hmac_key else None


def signed_unpack(
    data: str | bytes | None,
    hmac_data: str | None,
    hmac_keys: Sequence[str] | None,
) -> dict[str, Any] | None:
    """Unpack data and check that it was signed with hmac_key.

    :param data: json string that was singed_packed.
    :param hmac_data: hmac data that was generated from json by hmac_key on
                      user side
    :param hmac_keys: server side hmac_keys, one of these should be the same
                      as user used to sign with

    :returns: None in case of something wrong, Object in case of everything OK.
    """
    # NOTE(boris-42): For security reason, if there is no hmac_data or
    #                 hmac_keys we don't trust data => return None.
    if not hmac_keys or not hmac_data or not data:
        return None
    hmac_data = hmac_data.strip()
    if not hmac_data:
        return None
    for hmac_key in hmac_keys:
        try:
            user_hmac_data = generate_hmac(data, hmac_key)
        except Exception:  # noqa: S110
            pass
        else:
            if hmac.compare_digest(hmac_data, user_hmac_data):
                try:
                    contents: dict[str, Any] = json.loads(
                        binary_decode(base64.urlsafe_b64decode(data))
                    )
                    contents["hmac_key"] = hmac_key
                    return contents
                except Exception:
                    return None
    return None


def itersubclasses(
    cls: type[_C], _seen: set[type] | None = None
) -> Generator[type[_C], None, None]:
    """Generator over all subclasses of a given class in depth first order."""

    _seen = _seen or set()
    try:
        subs = cls.__subclasses__()
    except TypeError:  # fails only when cls is type
        subs = cls.__subclasses__(cls)  # type: ignore[call-arg]
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub


def import_modules_from_package(package: str) -> None:
    """Import modules from package and append into sys.modules

    :param: package - Full package name. For example: rally.deploy.engines
    """
    path_parts = [os.path.dirname(__file__), ".."] + package.split(".")
    path = os.path.join(*path_parts)
    for root, dirs, files in os.walk(path):
        for filename in files:
            if filename.startswith("__") or not filename.endswith(".py"):
                continue
            new_package = ".".join(root.split(os.sep)).split("....")[1]
            module_name = f"{new_package}.{filename[:-3]}"
            __import__(module_name)


def shorten_id(span_id: str | int) -> int:
    """Convert from uuid4 to 64 bit id for OpenTracing"""
    int64_max = (1 << 64) - 1
    if isinstance(span_id, int):
        return span_id & int64_max
    try:
        short_id = uuid.UUID(span_id).int & int64_max
    except ValueError:
        # Return a new short id for this
        short_id = shorten_id(uuidutils.generate_uuid())
    return short_id


def uuid_to_int128(span_uuid: str | int) -> int:
    """Convert from uuid4 to 128 bit id for OpenTracing"""
    if isinstance(span_uuid, int):
        return span_uuid
    try:
        span_int = uuid.UUID(span_uuid).int
    except ValueError:
        # Return a new short id for this
        span_int = uuid_to_int128(uuidutils.generate_uuid())
    return span_int



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import json
import os
from typing import Any

from oslo_utils import encodeutils
from oslo_utils import uuidutils
import prettytable

from osprofiler.cmd import cliutils
from osprofiler.drivers import base
from osprofiler import exc


class BaseCommand:
    group_name: str | None = None


class TraceCommands(BaseCommand):
    group_name = "trace"

    @cliutils.arg("trace", help="File with trace or trace id")
    @cliutils.arg(
        "--connection-string",
        dest="conn_str",
        default=(cliutils.env("OSPROFILER_CONNECTION_STRING")),
        help="Storage driver's connection string. Defaults to "
        "env[OSPROFILER_CONNECTION_STRING] if set",
    )
    @cliutils.arg(
        "--transport-url",
        dest="transport_url",
        help="Oslo.messaging transport URL (for messaging:// driver "
        "only), e.g. rabbit://user:password@host:5672/",
    )
    @cliutils.arg(
        "--idle-timeout",
        dest="idle_timeout",
        type=int,
        default=1,
        help="How long to wait for the trace to finish, in seconds "
        "(for messaging:// driver only)",
    )
    @cliutils.arg(
        "--json",
        dest="use_json",
        action="store_true",
        help="show trace in JSON",
    )
    @cliutils.arg(
        "--html",
        dest="use_html",
        action="store_true",
        help="show trace in HTML",
    )
    @cliutils.arg(
        "--local-libs",
        dest="local_libs",
        action="store_true",
        help="use local static files of html in /libs/",
    )
    @cliutils.arg(
        "--dot",
        dest="use_dot",
        action="store_true",
        help="show trace in DOT language",
    )
    @cliutils.arg(
        "--render-dot",
        dest="render_dot_filename",
        help="filename for rendering the dot graph in pdf format",
    )
    @cliutils.arg("--out", dest="file_name", help="save output in file")
    def show(self, args: argparse.Namespace) -> None:
        """Display trace results in HTML, JSON or DOT format."""

        if not args.conn_str:
            raise exc.CommandError(
                "You must provide connection string via"
                " either --connection-string or "
                "via env[OSPROFILER_CONNECTION_STRING]"
            )

        trace = None

        if not uuidutils.is_uuid_like(args.trace):
            trace = json.load(open(args.trace))
        else:
            try:
                engine = base.get_driver(args.conn_str, **args.__dict__)
            except Exception as e:
                raise exc.CommandError(str(e))

            trace = engine.get_report(args.trace)

        if not trace or not trace.get("children"):
            msg = (
                f"Trace with UUID {args.trace} not found. Please check the "
                f"HMAC key used in the command."
            )
            raise exc.CommandError(msg)

        # Since datetime.datetime is not JSON serializable by default,
        # this method will handle that.
        def datetime_json_serialize(obj: Any) -> Any:
            if hasattr(obj, "isoformat"):
                return obj.isoformat()
            else:
                return obj

        if args.use_json:
            output = json.dumps(
                trace,
                default=datetime_json_serialize,
                separators=(",", ": "),
                indent=2,
            )
        elif args.use_html:
            with open(
                os.path.join(os.path.dirname(__file__), "template.html")
            ) as html_template:
                output = html_template.read().replace(
                    "$DATA",
                    json.dumps(
                        trace,
                        indent=4,
                        separators=(",", ": "),
                        default=datetime_json_serialize,
                    ),
                )
                if args.local_libs:
                    output = output.replace("$LOCAL", "true")
                else:
                    output = output.replace("$LOCAL", "false")
        elif args.use_dot:
            dot_graph = self._create_dot_graph(trace)
            output = dot_graph.source
            if args.render_dot_filename:
                dot_graph.render(args.render_dot_filename, cleanup=True)
        else:
            raise exc.CommandError(
                "You should choose one of the following "
                "output formats: json, html or dot."
            )

        if args.file_name:
            with open(args.file_name, "w+") as output_file:
                output_file.write(output)
        else:
            print(output)

    def _create_dot_graph(self, trace: dict[str, Any]) -> Any:
        try:
            import graphviz  # type: ignore[import-not-found]
        except ImportError:
            raise exc.CommandError(
                "graphviz library is required to use this option."
            )

        dot = graphviz.Digraph(format="pdf")
        next_id = [0]

        def _create_node(info: dict[str, Any]) -> str:
            time_taken = info["finished"] - info["started"]
            service = info["service"] + ":" if "service" in info else ""
            name = info["name"]
            label = f"{service}{name} - {time_taken} ms"

            if name == "wsgi":
                req = info["meta.raw_payload.wsgi-start"]["info"]["request"]
                label = "{}\\n{} {}..".format(
                    label, req["method"], req["path"][:30]
                )
            elif name == "rpc" or name == "driver":
                raw = info[f"meta.raw_payload.{name}-start"]
                fn_name = raw["info"]["function"]["name"]
                label = "{}\\n{}".format(label, fn_name.split(".")[-1])

            node_id = str(next_id[0])
            next_id[0] += 1
            dot.node(node_id, label)
            return node_id

        def _create_sub_graph(root: dict[str, Any]) -> str:
            rid = _create_node(root["info"])
            for child in root["children"]:
                cid = _create_sub_graph(child)
                dot.edge(rid, cid)
            return rid

        _create_sub_graph(trace)
        return dot

    @cliutils.arg(
        "--connection-string",
        dest="conn_str",
        default=cliutils.env("OSPROFILER_CONNECTION_STRING"),
        help="Storage driver's connection string. Defaults to "
        "env[OSPROFILER_CONNECTION_STRING] if set",
    )
    @cliutils.arg(
        "--error-trace",
        dest="error_trace",
        type=bool,
        default=False,
        help="List all traces that contain error.",
    )
    def list(self, args: argparse.Namespace) -> None:
        """List all traces"""
        if not args.conn_str:
            raise exc.CommandError(
                "You must provide connection string via"
                " either --connection-string or "
                "via env[OSPROFILER_CONNECTION_STRING]"
            )
        try:
            engine = base.get_driver(args.conn_str, **args.__dict__)
        except Exception as e:
            raise exc.CommandError(str(e))

        fields = ("base_id", "timestamp")
        pretty_table = prettytable.PrettyTable(fields)
        pretty_table.align = "l"
        if not args.error_trace:
            traces = engine.list_traces(set(fields))
        else:
            traces = engine.list_error_traces()
        for trace in traces:
            row = [trace[field] for field in fields]
            pretty_table.add_row(row)
        print(encodeutils.safe_encode(pretty_table.get_string()).decode())



# Copyright (c) 2016 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Classes to use VMware vRealize Log Insight as the trace data store.
"""

import json
import logging as log
from typing import Any
from urllib import parse as urlparse

import netaddr
from oslo_concurrency.lockutils import synchronized
import requests

from osprofiler.drivers import base
from osprofiler import exc

LOG = log.getLogger(__name__)


class LogInsightDriver(base.Driver):
    """Driver for storing trace data in VMware vRealize Log Insight.

    The driver uses Log Insight ingest service to store trace data and uses
    the query service to retrieve it. The minimum required Log Insight version
    is 3.3.

    The connection string to initialize the driver should be of the format:
    loginsight://<username>:<password>@<loginsight-host>

    If the username or password contains the character ':' or '@', it must be
    escaped using URL encoding. For example, the connection string to connect
    to Log Insight server at 10.1.2.3 using username "osprofiler" and password
    "p@ssword" is: loginsight://osprofiler:p%40ssword@10.1.2.3
    """

    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            connection_str, project=project, service=service, host=host
        )

        parsed_connection = urlparse.urlparse(connection_str)
        try:
            creds, host = parsed_connection.netloc.split("@")
            username, password = creds.split(":")
        except ValueError:
            raise ValueError(
                "Connection string format is: loginsight://"
                "<username>:<password>@<loginsight-host>. If the "
                "username or password contains the character '@' "
                "or ':', it must be escaped using URL encoding."
            )

        username = urlparse.unquote(username)
        password = urlparse.unquote(password)
        self._client = LogInsightClient(host, username, password)

        self._client.login()

    @classmethod
    def get_name(cls) -> str:
        return "loginsight"

    def notify(self, info: dict[str, Any], **kwargs: Any) -> None:
        """Send trace to Log Insight server."""

        trace = info.copy()
        trace["project"] = self.project
        trace["service"] = self.service

        event: dict[str, Any] = {"text": "OSProfiler trace"}

        def _create_field(name: str, content: Any) -> dict[str, Any]:
            return {"name": name, "content": content}

        event["fields"] = [
            _create_field("base_id", trace["base_id"]),
            _create_field("trace_id", trace["trace_id"]),
            _create_field("project", trace["project"]),
            _create_field("service", trace["service"]),
            _create_field("name", trace["name"]),
            _create_field("trace", json.dumps(trace)),
        ]

        self._client.send_event(event)

    def get_report(self, base_id: str) -> dict[str, Any]:
        """Retrieves and parses trace data from Log Insight.

        :param base_id: Trace base ID
        """
        response = self._client.query_events({"base_id": base_id})

        if "events" in response:
            for event in response["events"]:
                if "fields" not in event:
                    continue

                for field in event["fields"]:
                    if field["name"] == "trace":
                        trace = json.loads(field["content"])
                        trace_id = trace["trace_id"]
                        parent_id = trace["parent_id"]
                        name = trace["name"]
                        project = trace["project"]
                        service = trace["service"]
                        host = trace["info"]["host"]
                        timestamp = trace["timestamp"]

                        self._append_results(
                            trace_id,
                            parent_id,
                            name,
                            project,
                            service,
                            host,
                            timestamp,
                            trace,
                        )
                        break

        return self._parse_results()


class LogInsightClient:
    """A minimal Log Insight client."""

    LI_OSPROFILER_AGENT_ID = "F52D775B-6017-4787-8C8A-F21AE0AEC057"

    # API paths
    SESSIONS_PATH = "api/v1/sessions"
    CURRENT_SESSIONS_PATH = "api/v1/sessions/current"
    EVENTS_INGEST_PATH = f"api/v1/events/ingest/{LI_OSPROFILER_AGENT_ID}"
    QUERY_EVENTS_BASE_PATH = "api/v1/events"

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        api_port: int = 9000,
        api_ssl_port: int = 9543,
        query_timeout: int = 60000,
    ) -> None:
        self._host = host
        self._username = username
        self._password = password
        self._api_port = api_port
        self._api_ssl_port = api_ssl_port
        self._query_timeout = query_timeout
        self._session = requests.Session()
        self._session_id: str | None = None

    def _build_base_url(self, scheme: str) -> str:
        proto_str = f"{scheme}://"
        host_str = (
            f"[{self._host}]" if netaddr.valid_ipv6(self._host) else self._host
        )
        port_str = ":%s" % (
            self._api_ssl_port if scheme == "https" else self._api_port
        )
        return proto_str + host_str + port_str

    def _check_response(self, resp: requests.Response) -> None:
        if resp.status_code == 440:
            raise exc.LogInsightLoginTimeout()

        if not resp.ok:
            msg = "n/a"
            if resp.text:
                try:
                    body = json.loads(resp.text)
                    msg = body.get("errorMessage", msg)
                except ValueError:
                    pass
            else:
                msg = resp.reason or msg
            raise exc.LogInsightAPIError(msg)

    def _send_request(
        self,
        method: str,
        scheme: str,
        path: str,
        headers: dict[str, str] | None = None,
        body: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        url = f"{self._build_base_url(scheme)}/{path}"

        headers = headers or {}
        headers["content-type"] = "application/json"
        body = body or {}
        params = params or {}

        req = requests.Request(
            method, url, headers=headers, data=json.dumps(body), params=params
        )
        prepped = req.prepare()
        resp = self._session.send(prepped, verify=False)

        self._check_response(resp)
        return resp.json()

    def _get_auth_header(self) -> dict[str, str]:
        return {"X-LI-Session-Id": self._session_id or ""}

    def _trunc_session_id(self) -> str | None:
        if self._session_id:
            return self._session_id[-5:]
        return None

    def _is_current_session_active(self) -> bool:
        try:
            self._send_request(
                "get",
                "https",
                self.CURRENT_SESSIONS_PATH,
                headers=self._get_auth_header(),
            )
            LOG.debug(
                "Current session %s is active.", self._trunc_session_id()
            )
            return True
        except (exc.LogInsightLoginTimeout, exc.LogInsightAPIError):
            LOG.debug(
                "Current session %s is not active.", self._trunc_session_id()
            )
            return False

    @synchronized("li_login_lock")
    def login(self) -> None:
        # Another thread might have created the session while the current
        # thread was waiting for the lock.
        if self._session_id and self._is_current_session_active():
            return

        LOG.info("Logging into Log Insight server: %s.", self._host)
        resp = self._send_request(
            "post",
            "https",
            self.SESSIONS_PATH,
            body={"username": self._username, "password": self._password},
        )

        self._session_id = resp["sessionId"]
        LOG.debug("Established session %s.", self._trunc_session_id())

    def send_event(self, event: dict[str, Any]) -> None:
        events = {"events": [event]}
        self._send_request(
            "post", "http", self.EVENTS_INGEST_PATH, body=events
        )

    def query_events(self, params: dict[str, str]) -> Any:
        # Assumes that the keys and values in the params are strings and
        # the operator is "CONTAINS".
        constraints = []
        for field, value in params.items():
            constraints.append(f"{field}/CONTAINS+{value}")
        constraints.append("timestamp/GT+0")

        path = "{}/{}".format(
            self.QUERY_EVENTS_BASE_PATH, "/".join(constraints)
        )

        def _query_events() -> Any:
            return self._send_request(
                "get",
                "https",
                path,
                headers=self._get_auth_header(),
                params={"limit": 20000, "timeout": self._query_timeout},
            )

        try:
            resp = _query_events()
        except exc.LogInsightLoginTimeout:
            # Login again and re-try.
            LOG.debug("Current session timed out.")
            self.login()
            resp = _query_events()

        return resp



# Copyright 2016 Mirantis Inc.
# Copyright 2016 IBM Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import Any
from unittest import mock

from oslo_serialization import jsonutils

from osprofiler.drivers.redis_driver import Redis
from osprofiler.tests import test


class RedisParserTestCase(test.TestCase):
    def setUp(self):
        super().setUp()
        self.redisdb = Redis("redis://localhost:6379")

    def test_build_empty_tree(self):
        self.assertEqual([], self.redisdb._build_tree({}))

    def test_build_complex_tree(self):
        test_input = {
            "2": {"parent_id": "0", "trace_id": "2", "info": {"started": 1}},
            "1": {"parent_id": "0", "trace_id": "1", "info": {"started": 0}},
            "21": {"parent_id": "2", "trace_id": "21", "info": {"started": 6}},
            "22": {"parent_id": "2", "trace_id": "22", "info": {"started": 7}},
            "11": {"parent_id": "1", "trace_id": "11", "info": {"started": 1}},
            "113": {
                "parent_id": "11",
                "trace_id": "113",
                "info": {"started": 3},
            },
            "112": {
                "parent_id": "11",
                "trace_id": "112",
                "info": {"started": 2},
            },
            "114": {
                "parent_id": "11",
                "trace_id": "114",
                "info": {"started": 5},
            },
        }

        expected_output = [
            {
                "parent_id": "0",
                "trace_id": "1",
                "info": {"started": 0},
                "children": [
                    {
                        "parent_id": "1",
                        "trace_id": "11",
                        "info": {"started": 1},
                        "children": [
                            {
                                "parent_id": "11",
                                "trace_id": "112",
                                "info": {"started": 2},
                                "children": [],
                            },
                            {
                                "parent_id": "11",
                                "trace_id": "113",
                                "info": {"started": 3},
                                "children": [],
                            },
                            {
                                "parent_id": "11",
                                "trace_id": "114",
                                "info": {"started": 5},
                                "children": [],
                            },
                        ],
                    }
                ],
            },
            {
                "parent_id": "0",
                "trace_id": "2",
                "info": {"started": 1},
                "children": [
                    {
                        "parent_id": "2",
                        "trace_id": "21",
                        "info": {"started": 6},
                        "children": [],
                    },
                    {
                        "parent_id": "2",
                        "trace_id": "22",
                        "info": {"started": 7},
                        "children": [],
                    },
                ],
            },
        ]

        result = self.redisdb._build_tree(test_input)
        self.assertEqual(expected_output, result)

    def test_get_report_empty(self):
        self.redisdb.db = mock.MagicMock()
        self.redisdb.db.scan_iter.return_value = []

        expected = {
            "info": {
                "name": "total",
                "started": 0,
                "finished": None,
                "last_trace_started": None,
            },
            "children": [],
            "stats": {},
        }

        base_id = "10"
        self.assertEqual(expected, self.redisdb.get_report(base_id))

    def test_get_report(self):
        self.redisdb.db = mock.MagicMock()
        result_elements: list[dict[str, Any]] = [
            {
                "info": {
                    "project": None,
                    "host": "ubuntu",
                    "request": {
                        "path": "/v2/a322b5049d224a90bf8786c644409400/volumes",
                        "scheme": "http",
                        "method": "POST",
                        "query": "",
                    },
                    "service": None,
                },
                "name": "wsgi-start",
                "service": "main",
                "timestamp": "2015-12-23T14:02:22.338776",
                "trace_id": "06320327-2c2c-45ae-923a-515de890276a",
                "project": "keystone",
                "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
            },
            {
                "info": {"project": None, "host": "ubuntu", "service": None},
                "name": "wsgi-stop",
                "service": "main",
                "timestamp": "2015-12-23T14:02:22.380405",
                "trace_id": "839ca3f1-afcb-45be-a4a1-679124c552bf",
                "project": "keystone",
                "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
            },
            {
                "info": {
                    "project": None,
                    "host": "ubuntu",
                    "db": {"params": {}, "statement": "SELECT 1"},
                    "service": None,
                },
                "name": "db-start",
                "service": "main",
                "timestamp": "2015-12-23T14:02:22.395365",
                "trace_id": "1baf1d24-9ca9-4f4c-bd3f-01b7e0c0735a",
                "project": "keystone",
                "parent_id": "06320327-2c2c-45ae-923a-515de890276a",
                "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
            },
            {
                "info": {"project": None, "host": "ubuntu", "service": None},
                "name": "db-stop",
                "service": "main",
                "timestamp": "2015-12-23T14:02:22.415486",
                "trace_id": "1baf1d24-9ca9-4f4c-bd3f-01b7e0c0735a",
                "project": "keystone",
                "parent_id": "06320327-2c2c-45ae-923a-515de890276a",
                "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
            },
            {
                "info": {
                    "project": None,
                    "host": "ubuntu",
                    "request": {
                        "path": "/v2/a322b5049d224a90bf8786c644409400/volumes",
                        "scheme": "http",
                        "method": "GET",
                        "query": "",
                    },
                    "service": None,
                },
                "name": "wsgi-start",
                "service": "main",
                "timestamp": "2015-12-23T14:02:22.427444",
                "trace_id": "016c97fd-87f3-40b2-9b55-e431156b694b",
                "project": "keystone",
                "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
            },
        ]
        results = {
            result["base_id"]
            + "_"
            + result["trace_id"]
            + "_"
            + result["timestamp"]: result
            for result in result_elements
        }

        expected = {
            "children": [
                {
                    "children": [
                        {
                            "children": [],
                            "info": {
                                "finished": 76,
                                "host": "ubuntu",
                                "meta.raw_payload.db-start": {
                                    "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",  # noqa: E501
                                    "info": {
                                        "db": {
                                            "params": {},
                                            "statement": "SELECT 1",
                                        },
                                        "host": "ubuntu",
                                        "project": None,
                                        "service": None,
                                    },
                                    "name": "db-start",
                                    "parent_id": "06320327-2c2c-45ae-923a-515de890276a",  # noqa: E501
                                    "project": "keystone",
                                    "service": "main",
                                    "timestamp": "2015-12-23T14:02:22.395365",
                                    "trace_id": "1baf1d24-9ca9-4f4c-bd3f-01b7e0c0735a",  # noqa: E501
                                },
                                "meta.raw_payload.db-stop": {
                                    "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",  # noqa: E501
                                    "info": {
                                        "host": "ubuntu",
                                        "project": None,
                                        "service": None,
                                    },
                                    "name": "db-stop",
                                    "parent_id": "06320327-2c2c-45ae-923a-515de890276a",  # noqa: E501
                                    "project": "keystone",
                                    "service": "main",
                                    "timestamp": "2015-12-23T14:02:22.415486",
                                    "trace_id": "1baf1d24-9ca9-4f4c-bd3f-01b7e0c0735a",  # noqa: E501
                                },
                                "name": "db",
                                "project": "keystone",
                                "service": "main",
                                "started": 56,
                                "exception": "None",
                            },
                            "parent_id": "06320327-2c2c-45ae-923a-515de890276a",  # noqa: E501
                            "trace_id": "1baf1d24-9ca9-4f4c-bd3f-01b7e0c0735a",
                        }
                    ],
                    "info": {
                        "finished": 0,
                        "host": "ubuntu",
                        "meta.raw_payload.wsgi-start": {
                            "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                            "info": {
                                "host": "ubuntu",
                                "project": None,
                                "request": {
                                    "method": "POST",
                                    "path": "/v2/a322b5049d224a90bf8"
                                    "786c644409400/volumes",
                                    "query": "",
                                    "scheme": "http",
                                },
                                "service": None,
                            },
                            "name": "wsgi-start",
                            "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",  # noqa: E501
                            "project": "keystone",
                            "service": "main",
                            "timestamp": "2015-12-23T14:02:22.338776",
                            "trace_id": "06320327-2c2c-45ae-923a-515de890276a",
                        },
                        "name": "wsgi",
                        "project": "keystone",
                        "service": "main",
                        "started": 0,
                    },
                    "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                    "trace_id": "06320327-2c2c-45ae-923a-515de890276a",
                },
                {
                    "children": [],
                    "info": {
                        "finished": 41,
                        "host": "ubuntu",
                        "meta.raw_payload.wsgi-stop": {
                            "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                            "info": {
                                "host": "ubuntu",
                                "project": None,
                                "service": None,
                            },
                            "name": "wsgi-stop",
                            "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",  # noqa: E501
                            "project": "keystone",
                            "service": "main",
                            "timestamp": "2015-12-23T14:02:22.380405",
                            "trace_id": "839ca3f1-afcb-45be-a4a1-679124c552bf",
                        },
                        "name": "wsgi",
                        "project": "keystone",
                        "service": "main",
                        "started": 41,
                        "exception": "None",
                    },
                    "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                    "trace_id": "839ca3f1-afcb-45be-a4a1-679124c552bf",
                },
                {
                    "children": [],
                    "info": {
                        "finished": 88,
                        "host": "ubuntu",
                        "meta.raw_payload.wsgi-start": {
                            "base_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                            "info": {
                                "host": "ubuntu",
                                "project": None,
                                "request": {
                                    "method": "GET",
                                    "path": "/v2/a322b5049d224a90bf"
                                    "8786c644409400/volumes",
                                    "query": "",
                                    "scheme": "http",
                                },
                                "service": None,
                            },
                            "name": "wsgi-start",
                            "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",  # noqa: E501
                            "project": "keystone",
                            "service": "main",
                            "timestamp": "2015-12-23T14:02:22.427444",
                            "trace_id": "016c97fd-87f3-40b2-9b55-e431156b694b",
                        },
                        "name": "wsgi",
                        "project": "keystone",
                        "service": "main",
                        "started": 88,
                    },
                    "parent_id": "7253ca8c-33b3-4f84-b4f1-f5a4311ddfa4",
                    "trace_id": "016c97fd-87f3-40b2-9b55-e431156b694b",
                },
            ],
            "info": {
                "finished": 88,
                "name": "total",
                "started": 0,
                "last_trace_started": 88,
            },
            "stats": {
                "db": {"count": 1, "duration": 20},
                "wsgi": {"count": 3, "duration": 0},
            },
        }

        self.redisdb.db.scan_iter.return_value = list(results.keys())

        def side_effect(*args, **kwargs):
            return jsonutils.dumps(results[args[0]])

        self.redisdb.db.get.side_effect = side_effect

        base_id = "10"

        result = self.redisdb.get_report(base_id)

        expected_filter = self.redisdb.namespace + "10*"
        self.redisdb.db.scan_iter.assert_called_once_with(
            match=expected_filter
        )
        self.assertEqual(expected, result)



# Copyright 2016 Mirantis Inc.
# Copyright 2016 IBM Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Generator
from typing import Any, cast
from urllib import parse as parser

from oslo_config import cfg
from oslo_serialization import jsonutils

from osprofiler.drivers import base
from osprofiler import exc


class Redis(base.Driver):
    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        conf: cfg.ConfigOpts = cfg.CONF,
        **kwargs: Any,
    ) -> None:
        """Redis driver for OSProfiler."""

        super().__init__(
            connection_str,
            project=project,
            service=service,
            host=host,
            conf=conf,
            **kwargs,
        )
        try:
            from redis import Redis as _Redis
        except ImportError:
            raise exc.CommandError(
                "To use OSProfiler with Redis driver, "
                "please install `redis` library. "
                "To install with pip:\n `pip install redis`."
            )

        # only connection over network is supported with schema
        # redis://[:password]@host[:port][/db]
        self.db = _Redis.from_url(self.connection_str)
        self.namespace_opt = "osprofiler_opt:"
        self.namespace = "osprofiler:"  # legacy
        self.namespace_error = "osprofiler_error:"

    @classmethod
    def get_name(cls) -> str:
        return "redis"

    def notify(self, info: dict[str, Any], **kwargs: Any) -> None:
        """Send notifications to Redis.

        :param info:  Contains information about trace element.
                      In payload dict there are always 3 ids:
                      "base_id" - uuid that is common for all notifications
                      related to one trace. Used to simplify retrieving of all
                      trace elements from Redis.
                      "parent_id" - uuid of parent element in trace
                      "trace_id" - uuid of current element in trace
                      With parent_id and trace_id it's quite simple to build
                      tree of trace elements, which simplify analyze of trace.
        """
        data = info.copy()
        data["project"] = self.project
        data["service"] = self.service
        key = self.namespace_opt + data["base_id"]
        self.db.lpush(key, jsonutils.dumps(data))

        if (
            self.filter_error_trace
            and data.get("info", {}).get("etype") is not None
        ):
            self.notify_error_trace(data)

    def notify_error_trace(self, data: dict[str, Any]) -> None:
        """Store base_id and timestamp of error trace to a separate key."""
        key = self.namespace_error + data["base_id"]
        value = jsonutils.dumps(
            {"base_id": data["base_id"], "timestamp": data["timestamp"]}
        )
        self.db.set(key, value)

    def list_traces(
        self, fields: set[str] | None = None
    ) -> list[dict[str, Any]]:
        """Query all traces from the storage.

        :param fields: Set of trace fields to return. Defaults to 'base_id'
                       and 'timestamp'
        :returns: List of traces, where each trace is a dictionary containing
                  at least `base_id` and `timestamp`.
        """
        fields = set(fields or self.default_trace_fields)

        # first get legacy events
        result = self._list_traces_legacy(fields)

        # with optimized schema trace events are stored in a list
        ids = self.db.scan_iter(match=self.namespace_opt + "*")
        for i in ids:
            # for each trace query the first event to have a timestamp
            raw = cast(bytes | None, self.db.lindex(i, 1))
            if raw is None:
                continue
            first_event = jsonutils.loads(raw)
            result.append(
                {
                    key: value
                    for key, value in first_event.items()
                    if key in fields
                }
            )
        return result

    def _list_traces_legacy(self, fields: set[str]) -> list[dict[str, Any]]:
        # With current schema every event is stored under its own unique key
        # To query all traces we first need to get all keys, then
        # get all events, sort them and pick up only the first one
        ids = self.db.scan_iter(match=self.namespace + "*")
        traces = [
            jsonutils.loads(raw)
            for i in ids
            if (raw := cast(bytes | None, self.db.get(i))) is not None
        ]
        traces.sort(key=lambda x: x["timestamp"])
        seen_ids: set[str] = set()
        result: list[dict[str, Any]] = []
        for trace in traces:
            if trace["base_id"] not in seen_ids:
                seen_ids.add(trace["base_id"])
                result.append(
                    {
                        key: value
                        for key, value in trace.items()
                        if key in fields
                    }
                )
        return result

    def list_error_traces(self) -> list[dict[str, Any]]:
        """Returns all traces that have error/exception."""
        ids = self.db.scan_iter(match=self.namespace_error + "*")
        traces = [
            jsonutils.loads(raw)
            for i in ids
            if (raw := cast(bytes | None, self.db.get(i))) is not None
        ]
        traces.sort(key=lambda x: x["timestamp"])
        seen_ids: set[str] = set()
        result: list[dict[str, Any]] = []
        for trace in traces:
            if trace["base_id"] not in seen_ids:
                seen_ids.add(trace["base_id"])
                result.append(trace)

        return result

    def get_report(self, base_id: str) -> dict[str, Any]:
        """Retrieves and parses notification from Redis.

        :param base_id: Base id of trace elements.
        """

        def iterate_events() -> Generator[bytes, None, None]:
            for key in self.db.scan_iter(
                match=self.namespace + base_id + "*"
            ):  # legacy
                data = cast(bytes | None, self.db.get(key))
                if data is not None:
                    yield data

            yield from cast(
                list[bytes],
                self.db.lrange(self.namespace_opt + base_id, 0, -1),
            )

        for data in iterate_events():
            n = jsonutils.loads(data)
            trace_id = n["trace_id"]
            parent_id = n["parent_id"]
            name = n["name"]
            project = n["project"]
            service = n["service"]
            host = n["info"]["host"]
            timestamp = n["timestamp"]

            self._append_results(
                trace_id, parent_id, name, project, service, host, timestamp, n
            )

        return self._parse_results()


class RedisSentinel(Redis, base.Driver):
    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        conf: cfg.ConfigOpts = cfg.CONF,
        **kwargs: Any,
    ) -> None:
        """Redis driver for OSProfiler."""

        super().__init__(
            connection_str,
            project=project,
            service=service,
            host=host,
            conf=conf,
            **kwargs,
        )
        try:
            from redis.sentinel import Sentinel
        except ImportError:
            raise exc.CommandError(
                "To use this command, you should install "
                "'redis' manually. Use command:\n "
                "'pip install redis'."
            )

        self.conf = conf
        socket_timeout = self.conf.profiler.socket_timeout
        parsed_url = parser.urlparse(self.connection_str)
        sentinel = Sentinel(  # type: ignore[no-untyped-call]
            [(parsed_url.hostname, int(parsed_url.port))],  # type: ignore[arg-type]
            password=parsed_url.password,
            socket_timeout=socket_timeout,
        )
        self.db = sentinel.master_for(  # type: ignore[no-untyped-call]
            self.conf.profiler.sentinel_service_name,
            socket_timeout=socket_timeout,
        )

    @classmethod
    def get_name(cls) -> str:
        return "redissentinel"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Guidelines for writing new hacking checks

 - Use only for OSProfiler specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range N3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the N3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to tests/unit/test_hacking.py

"""

from collections.abc import Generator
import re
import tokenize
from typing import Any

from hacking import core  # type: ignore[import-not-found]

re_no_construct_dict = re.compile(r"\sdict\(\)")
re_no_construct_list = re.compile(r"\slist\(\)")
re_str_format = re.compile(
    r"""
%            # start of specifier
\(([^)]+)\)  # mapping key, in group 1
[#0 +\-]?    # optional conversion flag
(?:-?\d*)?   # optional minimum field width
(?:\.\d*)?   # optional precision
[hLl]?       # optional length modifier
[A-z%]       # conversion modifier
""",
    re.X,
)
re_raises = re.compile(r"\s:raise[^s] *.*$|\s:raises *:.*$|\s:raises *[^:]+$")


def _check_triple(line: str, i: int, char: str) -> bool:
    return i + 2 < len(line) and (
        char == line[i] == line[i + 1] == line[i + 2]
    )


@core.flake8ext  # type: ignore[untyped-decorator]
def check_quotes(
    logical_line: str, filename: str, noqa: bool = False
) -> Generator[tuple[int, str], None, None]:
    """Check that single quotation marks are not used

    N350
    """

    if noqa:
        return

    in_string = False
    in_multiline_string = False
    single_quotas_are_used = False

    i = 0
    while i < len(logical_line):
        char = logical_line[i]

        if in_string:
            if char == "\"":
                in_string = False
            if char == "\\":
                i += 1  # ignore next char

        elif in_multiline_string:
            if _check_triple(logical_line, i, "\""):
                i += 2  # skip next 2 chars
                in_multiline_string = False

        elif char == "#":
            break

        elif char == "'":
            single_quotas_are_used = True
            break

        elif char == "\"":
            if _check_triple(logical_line, i, "\""):
                in_multiline_string = True
                i += 3
                continue
            in_string = True

        i += 1

    if single_quotas_are_used:
        yield (i, "N350 Remove Single quotes")


@core.flake8ext  # type: ignore[untyped-decorator]
def check_no_constructor_data_struct(
    logical_line: str, filename: str, noqa: bool = False
) -> Generator[tuple[int, str], None, None]:
    """Check that data structs (lists, dicts) are declared using literals

    N351
    """
    if noqa:
        return

    match = re_no_construct_dict.search(logical_line)
    if match:
        yield (0, "N351 Remove dict() construct and use literal {}")
    match = re_no_construct_list.search(logical_line)
    if match:
        yield (0, "N351 Remove list() construct and use literal []")


@core.flake8ext  # type: ignore[untyped-decorator]
def check_dict_formatting_in_string(
    logical_line: str, tokens: Any, noqa: bool = False
) -> Generator[tuple[int, str], None, None]:
    """Check that strings do not use dict-formatting with a single replacement

    N352
    """
    if noqa:
        return

    current_string = ""
    in_string = False
    for token_type, text, start, end, line in tokens:
        if token_type == tokenize.STRING:
            if not in_string:
                current_string = ""
                in_string = True
            current_string += text.strip("\"")
        elif token_type == tokenize.OP:
            if not current_string:
                continue
            # NOTE(stpierre): The string formatting operator % has
            # lower precedence than +, so we assume that the logical
            # string has concluded whenever we hit an operator of any
            # sort. (Most operators don't work for strings anyway.)
            # Some string operators do have higher precedence than %,
            # though, so you can technically trick this check by doing
            # things like:
            #
            #     "%(foo)s" * 1 % {"foo": 1}
            #     "%(foo)s"[:] % {"foo": 1}
            #
            # It also will produce false positives if you use explicit
            # parenthesized addition for two strings instead of
            # concatenation by juxtaposition, e.g.:
            #
            #     ("%(foo)s" + "%(bar)s") % vals
            #
            # But if you do any of those things, then you deserve all
            # of the horrible things that happen to you, and probably
            # many more.
            in_string = False
            if text == "%":
                format_keys = set()
                for match in re_str_format.finditer(current_string):
                    format_keys.add(match.group(1))
                if len(format_keys) == 1:
                    yield (
                        0,
                        "N352 Do not use mapping key string formatting "
                        "with a single key",
                    )
            if text != ")":
                # NOTE(stpierre): You can have a parenthesized string
                # followed by %, so a closing paren doesn't obviate
                # the possibility for a substitution operator like
                # every other operator does.
                current_string = ""
        elif token_type in (tokenize.NL, tokenize.COMMENT):
            continue
        else:
            in_string = False
            if token_type == tokenize.NEWLINE:
                current_string = ""


@core.flake8ext  # type: ignore[untyped-decorator]
def check_raises(physical_line: str, filename: str) -> tuple[int, str] | None:
    """Check raises usage

    N354
    """

    ignored_files = [
        "./tests/unit/test_hacking.py",
        "./tests/hacking/checks.py",
    ]
    if filename not in ignored_files:
        if re_raises.search(physical_line):
            return (
                0,
                "N354 ':Please use ':raises Exception: conditions' "
                "in docstrings.",
            )
    return None



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from typing import Any
from urllib import parse as parser

from oslo_config import cfg
from oslo_serialization import jsonutils

from osprofiler import _utils as utils
from osprofiler.drivers import base
from osprofiler import exc


class OTLP(base.Driver):
    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        conf: cfg.ConfigOpts = cfg.CONF,
        **kwargs: Any,
    ) -> None:
        """OTLP driver using OTLP exporters."""

        super().__init__(
            connection_str,
            project=project,
            service=service,
            host=host,
            conf=conf,
            **kwargs,
        )
        try:
            from opentelemetry import trace as trace_api

            from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
                OTLPSpanExporter,
            )  # noqa
            from opentelemetry.sdk.resources import Resource
            from opentelemetry.sdk.trace.export import BatchSpanProcessor
            from opentelemetry.sdk.trace import TracerProvider

            self.trace_api = trace_api
        except ImportError:
            raise exc.CommandError(
                "To use OSProfiler with OTLP exporters, "
                "please install `opentelemetry-sdk` and "
                "opentelemetry-exporter-otlp libraries. "
                "To install with pip:\n `pip install opentelemetry-sdk "
                "opentelemetry-exporter-otlp`."
            )

        service_name = self._get_service_name(conf, project, service)
        resource = Resource(attributes={"service.name": service_name})

        parsed_url = parser.urlparse(connection_str)
        # TODO("sahid"): We also want to handle https scheme?
        parsed_url = parsed_url._replace(scheme="http")

        self.trace_api.set_tracer_provider(TracerProvider(resource=resource))
        self.tracer = self.trace_api.get_tracer(__name__)

        exporter = OTLPSpanExporter(f"{parsed_url.geturl()}/v1/traces")
        self.trace_api.get_tracer_provider().add_span_processor(  # type: ignore[attr-defined]
            BatchSpanProcessor(exporter)
        )

        self.spans: collections.deque[Any] = collections.deque()

    def _get_service_name(
        self, conf: cfg.ConfigOpts, project: str | None, service: str | None
    ) -> str:
        prefix = conf.profiler_otlp.service_name_prefix
        if prefix:
            return f"{prefix}-{project}-{service}"
        return f"{project}-{service}"

    @classmethod
    def get_name(cls) -> str:
        return "otlp"

    def _kind(self, name: str) -> Any:
        if "wsgi" in name:
            return self.trace_api.SpanKind.SERVER
        elif "db" in name or "http" in name or "api" in name:
            return self.trace_api.SpanKind.CLIENT
        return self.trace_api.SpanKind.INTERNAL

    def _name(self, payload: dict[str, Any]) -> str:
        info = payload["info"]
        if info.get("request"):
            return "WSGI_{}_{}".format(
                info["request"]["method"], info["request"]["path"]
            )
        elif info.get("db"):
            return "SQL_{}".format(
                info["db"]["statement"].split(' ', 1)[0].upper()
            )
        elif info.get("requests"):
            return "REQUESTS_{}_{}".format(
                info["requests"]["method"], info["requests"]["hostname"]
            )
        return str(payload["name"]).rstrip("-start")

    def notify(self, info: dict[str, Any], **kwargs: Any) -> None:
        payload = info
        if payload["name"].endswith("start"):
            ctx = None
            trace_id = utils.uuid_to_int128(payload["base_id"])
            if payload["base_id"] != payload["parent_id"]:
                # only non-root spans should have a parent span and context
                parent = self.trace_api.SpanContext(
                    trace_id=trace_id,
                    span_id=utils.shorten_id(payload["parent_id"]),
                    is_remote=False,
                    trace_flags=self.trace_api.TraceFlags(
                        self.trace_api.TraceFlags.SAMPLED
                    ),
                )

                ctx = self.trace_api.set_span_in_context(
                    self.trace_api.NonRecordingSpan(parent)
                )

            # OTLP Tracing span
            span = self.tracer.start_span(
                name=self._name(payload),
                kind=self._kind(payload['name']),
                attributes=self.create_span_tags(payload),
                context=ctx,
            )

            span._context = self.trace_api.SpanContext(  # type: ignore[attr-defined]
                trace_id=trace_id,
                span_id=utils.shorten_id(payload["trace_id"]),
                is_remote=span.context.is_remote,  # type: ignore[attr-defined]
                trace_flags=span.context.trace_flags,  # type: ignore[attr-defined]
                trace_state=span.context.trace_state,  # type: ignore[attr-defined]
            )

            self.spans.append(span)
        else:
            span = self.spans.pop()

            # Store result of db call and function call
            for call in ("db", "function"):
                if payload.get("info", {}).get(call, {}).get("result"):
                    span.set_attribute(
                        "result", payload["info"][call]["result"]
                    )
            # Store result of requests
            if payload.get("info", {}).get("requests"):
                span.set_attribute(
                    "status_code", payload["info"]["requests"]["status_code"]
                )
            # Span error tag and log
            if payload["info"].get("etype"):
                span.set_attribute("error", True)
                span.add_event(
                    "log",
                    {
                        "error.kind": payload["info"]["etype"],
                        "message": payload["info"]["message"],
                    },
                )
            span.end()

    def get_report(self, base_id: str) -> dict[str, Any]:
        return self._parse_results()

    def list_traces(
        self, fields: set[str] | None = None
    ) -> list[dict[str, Any]]:
        return []

    def list_error_traces(self) -> list[dict[str, Any]]:
        return []

    def create_span_tags(self, payload: dict[str, Any]) -> dict[str, Any]:
        """Create tags an OpenTracing compatible span.

        :param info: Information from OSProfiler trace.
        :returns tags: A dictionary contains standard tags
                       from OpenTracing sematic conventions,
                       and some other custom tags related to http, db calls.
        """
        tags: dict[str, Any] = {}
        info = payload["info"]

        if info.get("db"):
            # DB calls
            tags["db.statement"] = info["db"]["statement"]
            tags["db.params"] = jsonutils.dumps(info["db"]["params"])
        elif info.get("request"):
            # WSGI call
            tags["http.path"] = info["request"]["path"]
            tags["http.query"] = info["request"]["query"]
            tags["http.method"] = info["request"]["method"]
            tags["http.scheme"] = info["request"]["scheme"]
        elif info.get("requests"):
            # requests call
            tags["http.path"] = info["requests"]["path"]
            tags["http.query"] = info["requests"]["query"]
            tags["http.method"] = info["requests"]["method"]
            tags["http.scheme"] = info["requests"]["scheme"]
            tags["http.hostname"] = info["requests"]["hostname"]
            tags["http.port"] = info["requests"]["port"]
        elif info.get("function"):
            # RPC, function calls
            if "args" in info["function"]:
                tags["args"] = info["function"]["args"]
            if "kwargs" in info["function"]:
                tags["kwargs"] = info["function"]["kwargs"]
            tags["name"] = info["function"]["name"]

        return tags



# Copyright (c) 2016 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

import ddt

from osprofiler.drivers import loginsight
from osprofiler import exc
from osprofiler.tests import test


@ddt.ddt
class LogInsightDriverTestCase(test.TestCase):
    BASE_ID = "8d28af1e-acc0-498c-9890-6908e33eff5f"

    def setUp(self):
        super().setUp()
        self._client = mock.Mock(spec=loginsight.LogInsightClient)
        self._project = "cinder"
        self._service = "osapi_volume"
        self._host = "ubuntu"
        with mock.patch.object(
            loginsight, "LogInsightClient", return_value=self._client
        ):
            self._driver = loginsight.LogInsightDriver(
                "loginsight://username:password@host",
                project=self._project,
                service=self._service,
                host=self._host,
            )

    @mock.patch.object(loginsight, "LogInsightClient")
    def test_init(self, client_class):
        client = mock.Mock()
        client_class.return_value = client

        loginsight.LogInsightDriver("loginsight://username:password@host")
        client_class.assert_called_once_with("host", "username", "password")
        client.login.assert_called_once_with()

    @ddt.data(
        "loginsight://username@host",
        "loginsight://username:p@ssword@host",
        "loginsight://us:rname:password@host",
    )
    def test_init_with_invalid_connection_string(self, conn_str):
        self.assertRaises(ValueError, loginsight.LogInsightDriver, conn_str)

    @mock.patch.object(loginsight, "LogInsightClient")
    def test_init_with_special_chars_in_conn_str(self, client_class):
        client = mock.Mock()
        client_class.return_value = client

        loginsight.LogInsightDriver("loginsight://username:p%40ssword@host")
        client_class.assert_called_once_with("host", "username", "p@ssword")
        client.login.assert_called_once_with()

    def test_get_name(self):
        self.assertEqual("loginsight", self._driver.get_name())

    def _create_trace(
        self,
        name,
        timestamp,
        parent_id="8d28af1e-acc0-498c-9890-6908e33eff5f",
        base_id=BASE_ID,
        trace_id="e465db5c-9672-45a1-b90b-da918f30aef6",
    ):
        return {
            "parent_id": parent_id,
            "name": name,
            "base_id": base_id,
            "trace_id": trace_id,
            "timestamp": timestamp,
            "info": {"host": self._host},
        }

    def _create_start_trace(self):
        return self._create_trace("wsgi-start", "2016-10-04t11:50:21.902303")

    def _create_stop_trace(self):
        return self._create_trace("wsgi-stop", "2016-10-04t11:50:30.123456")

    @mock.patch("json.dumps")
    def test_notify(self, dumps):
        json_str = mock.sentinel.json_str
        dumps.return_value = json_str

        trace = self._create_stop_trace()
        self._driver.notify(trace)

        trace["project"] = self._project
        trace["service"] = self._service
        exp_event = {
            "text": "OSProfiler trace",
            "fields": [
                {"name": "base_id", "content": trace["base_id"]},
                {"name": "trace_id", "content": trace["trace_id"]},
                {"name": "project", "content": trace["project"]},
                {"name": "service", "content": trace["service"]},
                {"name": "name", "content": trace["name"]},
                {"name": "trace", "content": json_str},
            ],
        }
        self._client.send_event.assert_called_once_with(exp_event)

    @mock.patch.object(loginsight.LogInsightDriver, "_append_results")
    @mock.patch.object(loginsight.LogInsightDriver, "_parse_results")
    def test_get_report(self, parse_results, append_results):
        start_trace = self._create_start_trace()
        start_trace["project"] = self._project
        start_trace["service"] = self._service

        stop_trace = self._create_stop_trace()
        stop_trace["project"] = self._project
        stop_trace["service"] = self._service

        resp = {
            "events": [
                {
                    "text": "OSProfiler trace",
                    "fields": [
                        {"name": "trace", "content": json.dumps(start_trace)}
                    ],
                },
                {
                    "text": "OSProfiler trace",
                    "fields": [
                        {"name": "trace", "content": json.dumps(stop_trace)}
                    ],
                },
            ]
        }
        self._client.query_events = mock.Mock(return_value=resp)

        self._driver.get_report(self.BASE_ID)
        self._client.query_events.assert_called_once_with(
            {"base_id": self.BASE_ID}
        )
        append_results.assert_has_calls(
            [
                mock.call(
                    start_trace["trace_id"],
                    start_trace["parent_id"],
                    start_trace["name"],
                    start_trace["project"],
                    start_trace["service"],
                    start_trace["info"]["host"],
                    start_trace["timestamp"],
                    start_trace,
                ),
                mock.call(
                    stop_trace["trace_id"],
                    stop_trace["parent_id"],
                    stop_trace["name"],
                    stop_trace["project"],
                    stop_trace["service"],
                    stop_trace["info"]["host"],
                    stop_trace["timestamp"],
                    stop_trace,
                ),
            ]
        )
        parse_results.assert_called_once_with()


class LogInsightClientTestCase(test.TestCase):
    def setUp(self):
        super().setUp()
        self._host = "localhost"
        self._username = "username"
        self._password = "password"  # noqa: S105
        self._client = loginsight.LogInsightClient(
            self._host, self._username, self._password
        )
        self._client._session_id = "4ff800d1-3175-4b49-9209-39714ea56416"

    def test_check_response_login_timeout(self):
        resp = mock.Mock(status_code=440)
        self.assertRaises(
            exc.LogInsightLoginTimeout, self._client._check_response, resp
        )

    def test_check_response_api_error(self):
        resp = mock.Mock(status_code=401, ok=False)
        resp.text = json.dumps(
            {
                "errorMessage": "Invalid username or password.",
                "errorCode": "FIELD_ERROR",
            }
        )
        e = self.assertRaises(
            exc.LogInsightAPIError, self._client._check_response, resp
        )
        self.assertEqual("Invalid username or password.", str(e))

    @mock.patch("requests.Request")
    @mock.patch("json.dumps")
    @mock.patch.object(loginsight.LogInsightClient, "_check_response")
    def test_send_request(self, check_resp, json_dumps, request_class):
        req = mock.Mock()
        request_class.return_value = req
        prep_req = mock.sentinel.prep_req
        req.prepare = mock.Mock(return_value=prep_req)

        data = mock.sentinel.data
        json_dumps.return_value = data

        self._client._session = mock.Mock()
        resp = mock.Mock()
        self._client._session.send = mock.Mock(return_value=resp)
        resp_json = mock.sentinel.resp_json
        resp.json = mock.Mock(return_value=resp_json)

        header = {"X-LI-Session-Id": "foo"}
        body = mock.sentinel.body
        params = mock.sentinel.params
        ret = self._client._send_request(
            "get", "https", "api/v1/events", header, body, params
        )

        self.assertEqual(resp_json, ret)
        exp_headers = {
            "X-LI-Session-Id": "foo",
            "content-type": "application/json",
        }
        request_class.assert_called_once_with(
            "get",
            "https://localhost:9543/api/v1/events",
            headers=exp_headers,
            data=data,
            params=mock.sentinel.params,
        )
        self._client._session.send.assert_called_once_with(
            prep_req, verify=False
        )
        check_resp.assert_called_once_with(resp)

    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    def test_is_current_session_active_with_active_session(self, send_request):
        self.assertTrue(self._client._is_current_session_active())
        exp_header = {"X-LI-Session-Id": self._client._session_id}
        send_request.assert_called_once_with(
            "get", "https", "api/v1/sessions/current", headers=exp_header
        )

    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    def test_is_current_session_active_with_expired_session(
        self, send_request
    ):
        send_request.side_effect = exc.LogInsightLoginTimeout

        self.assertFalse(self._client._is_current_session_active())
        send_request.assert_called_once_with(
            "get",
            "https",
            "api/v1/sessions/current",
            headers={"X-LI-Session-Id": self._client._session_id},
        )

    @mock.patch.object(
        loginsight.LogInsightClient,
        "_is_current_session_active",
        return_value=True,
    )
    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    def test_login_with_current_session_active(
        self, send_request, is_current_session_active
    ):
        self._client.login()
        is_current_session_active.assert_called_once_with()
        send_request.assert_not_called()

    @mock.patch.object(
        loginsight.LogInsightClient,
        "_is_current_session_active",
        return_value=False,
    )
    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    def test_login(self, send_request, is_current_session_active):
        new_session_id = "569a80aa-be5c-49e5-82c1-bb62392d2667"
        resp = {"sessionId": new_session_id}
        send_request.return_value = resp

        self._client.login()
        is_current_session_active.assert_called_once_with()
        exp_body = {"username": self._username, "password": self._password}
        send_request.assert_called_once_with(
            "post", "https", "api/v1/sessions", body=exp_body
        )
        self.assertEqual(new_session_id, self._client._session_id)

    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    def test_send_event(self, send_request):
        event = mock.sentinel.event
        self._client.send_event(event)

        exp_body = {"events": [event]}
        exp_path = (
            f"api/v1/events/ingest/{self._client.LI_OSPROFILER_AGENT_ID}"
        )
        send_request.assert_called_once_with(
            "post", "http", exp_path, body=exp_body
        )

    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    def test_query_events(self, send_request):
        resp = mock.sentinel.response
        send_request.return_value = resp

        self.assertEqual(resp, self._client.query_events({"foo": "bar"}))
        exp_header = {"X-LI-Session-Id": self._client._session_id}
        exp_params = {"limit": 20000, "timeout": self._client._query_timeout}
        send_request.assert_called_once_with(
            "get",
            "https",
            "api/v1/events/foo/CONTAINS+bar/timestamp/GT+0",
            headers=exp_header,
            params=exp_params,
        )

    @mock.patch.object(loginsight.LogInsightClient, "_send_request")
    @mock.patch.object(loginsight.LogInsightClient, "login")
    def test_query_events_with_session_expiry(self, login, send_request):
        resp = mock.sentinel.response
        send_request.side_effect = [exc.LogInsightLoginTimeout, resp]

        self.assertEqual(resp, self._client.query_events({"foo": "bar"}))
        login.assert_called_once_with()
        exp_header = {"X-LI-Session-Id": self._client._session_id}
        exp_params = {"limit": 20000, "timeout": self._client._query_timeout}
        exp_send_request_call = mock.call(
            "get",
            "https",
            "api/v1/events/foo/CONTAINS+bar/timestamp/GT+0",
            headers=exp_header,
            params=exp_params,
        )
        send_request.assert_has_calls([exp_send_request_call] * 2)



# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import signal
import time
from typing import Any

from oslo_config import cfg
from oslo_utils import importutils

from osprofiler.drivers import base


class Messaging(base.Driver):
    def __init__(
        self,
        connection_str: str,
        project: str | None = None,
        service: str | None = None,
        host: str | None = None,
        context: Any = None,
        conf: cfg.ConfigOpts | None = None,
        transport_url: str | None = None,
        idle_timeout: int = 1,
        **kwargs: Any,
    ) -> None:
        """Driver that uses messaging as transport for notifications

        :param connection_str: OSProfiler driver connection string,
               equals to messaging://
        :param project: project name that will be included into notification
        :param service: service name that will be included into notification
        :param host: host name that will be included into notification
        :param context: oslo.messaging context
        :param conf: oslo.config CONF object
        :param transport_url: oslo.messaging transport, e.g.
               rabbit://rabbit:password@devstack:5672/
        :param idle_timeout: how long to wait for new notifications after
               the last one seen in the trace; this parameter is useful to
               collect full trace of asynchronous commands, e.g. when user
               runs `osprofiler` right after `openstack server create`
        :param kwargs: black hole for any other parameters
        """

        self.oslo_messaging = importutils.try_import("oslo_messaging")
        if not self.oslo_messaging:
            raise ValueError(
                "Oslo.messaging library is required for messaging driver"
            )

        super().__init__(
            connection_str, project=project, service=service, host=host
        )

        self.context = context

        if not conf:
            oslo_config = importutils.try_import("oslo_config")
            if not oslo_config:
                raise ValueError(
                    "Oslo.config library is required for messaging driver"
                )
            conf = oslo_config.cfg.CONF

        transport_kwargs: dict[str, Any] = {}
        if transport_url:
            transport_kwargs["url"] = transport_url

        self.transport = self.oslo_messaging.get_notification_transport(
            conf, **transport_kwargs
        )
        self.client = self.oslo_messaging.Notifier(
            self.transport,
            publisher_id=self.host,
            driver="messaging",
            topics=["profiler"],
            retry=0,
        )

        self.idle_timeout = idle_timeout

    @classmethod
    def get_name(cls) -> str:
        return "messaging"

    def notify(
        self, info: dict[str, Any], context: Any = None, **kwargs: Any
    ) -> None:
        """Send notifications to backend via oslo.messaging notifier API.

        :param info:  Contains information about trace element.
                      In payload dict there are always 3 ids:
                      "base_id" - uuid that is common for all notifications
                      related to one trace.
                      "parent_id" - uuid of parent element in trace
                      "trace_id" - uuid of current element in trace
                      With parent_id and trace_id it's quite simple to build
                      tree of trace elements, which simplify analyze of trace.

        :param context: request context that is mostly used to specify
                        current active user and tenant.
        """

        info["project"] = self.project
        info["service"] = self.service
        self.client.info(
            context or self.context,
            "profiler.{}".format(info["service"]),
            info,
        )

    def get_report(self, base_id: str) -> dict[str, Any]:
        notification_endpoint = NotifyEndpoint(self.oslo_messaging, base_id)
        endpoints = [notification_endpoint]
        targets = [self.oslo_messaging.Target(topic="profiler")]
        server = self.oslo_messaging.notify.get_notification_listener(
            self.transport, targets, endpoints, executor="threading"
        )

        state: dict[str, bool] = dict(running=False)
        sfn = functools.partial(signal_handler, state=state)

        # modify signal handlers to handle interruption gracefully
        old_sigterm_handler = signal.signal(signal.SIGTERM, sfn)
        old_sigint_handler = signal.signal(signal.SIGINT, sfn)

        try:
            server.start()
        except self.oslo_messaging.server.ServerListenError:
            # failed to start the server
            raise
        except SignalExit:
            print(
                "Execution interrupted while trying to connect to "
                "messaging server. No data was collected."
            )
            return {}

        # connected to server, now read the data
        try:
            # run until the trace is complete
            state["running"] = True

            while state["running"]:
                last_read_time = notification_endpoint.get_last_read_time()
                wait = self.idle_timeout - (time.time() - last_read_time)
                if wait < 0:
                    state["running"] = False
                else:
                    time.sleep(wait)
        except SignalExit:
            print("Execution interrupted. Terminating")
        finally:
            server.stop()
            server.wait()

        # restore original signal handlers
        signal.signal(signal.SIGTERM, old_sigterm_handler)
        signal.signal(signal.SIGINT, old_sigint_handler)

        events = notification_endpoint.get_messages()

        if not events:
            print(
                f"No events are collected for Trace UUID {base_id}. "
                f"Please note that osprofiler has read ALL events from "
                f"profiler topic, but has not found any for specified Trace "
                f"UUID."
            )

        for n in events:
            trace_id = n["trace_id"]
            parent_id = n["parent_id"]
            name = n["name"]
            project = n["project"]
            service = n["service"]
            host = n["info"]["host"]
            timestamp = n["timestamp"]

            self._append_results(
                trace_id, parent_id, name, project, service, host, timestamp, n
            )

        return self._parse_results()


class NotifyEndpoint:
    def __init__(self, oslo_messaging: Any, base_id: str) -> None:
        self.received_messages: list[Any] = []
        self.last_read_time = time.time()
        self.filter_rule = oslo_messaging.NotificationFilter(
            payload={"base_id": base_id}
        )

    def info(
        self,
        ctxt: Any,
        publisher_id: Any,
        event_type: Any,
        payload: Any,
        metadata: Any,
    ) -> None:
        self.received_messages.append(payload)
        self.last_read_time = time.time()

    def get_messages(self) -> list[Any]:
        return self.received_messages

    def get_last_read_time(self) -> float:
        return self.last_read_time  # time when the latest event was received


class SignalExit(BaseException):
    pass


def signal_handler(signum: Any, frame: Any, state: dict[str, bool]) -> None:
    state["running"] = False
    raise SignalExit()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import testtools

from osprofiler import initializer


class InitializerTestCase(testtools.TestCase):
    @mock.patch("osprofiler.notifier.set")
    @mock.patch("osprofiler.notifier.create")
    @mock.patch("osprofiler.web.enable")
    def test_initializer(
        self, web_enable_mock, notifier_create_mock, notifier_set_mock
    ):
        conf = mock.Mock()
        conf.profiler.connection_string = "driver://"
        conf.profiler.hmac_keys = "hmac_keys"
        context: dict[object, object] = {}
        project = "my-project"
        service = "my-service"
        host = "my-host"

        notifier_mock = mock.Mock()
        notifier_create_mock.return_value = notifier_mock

        initializer.init_from_conf(conf, context, project, service, host)

        notifier_create_mock.assert_called_once_with(
            "driver://",
            context=context,
            project=project,
            service=service,
            host=host,
            conf=conf,
        )
        notifier_set_mock.assert_called_once_with(notifier_mock)
        web_enable_mock.assert_called_once_with("hmac_keys")



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Callable
import os
from typing import Any


def env(*args: str, **kwargs: str) -> str:
    """Returns the first environment variable set.

    If all are empty, defaults to '' or keyword arg `default`.
    """
    for arg in args:
        value = os.environ.get(arg)
        if value:
            return value
    return kwargs.get("default", "")


def arg(
    *args: Any, **kwargs: Any
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator for CLI args.

    Example:

    >>> @arg("name", help="Name of the new entity")
        ... def entity_create(args):
        ... pass
    """

    def _decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        add_arg(func, *args, **kwargs)
        return func

    return _decorator


def add_arg(func: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
    """Bind CLI arguments to a shell.py `do_foo` function."""

    if not hasattr(func, "arguments"):
        setattr(func, "arguments", [])

    # NOTE(sirp): avoid dups that can occur when the module is shared across
    # tests.
    if (args, kwargs) not in getattr(func, "arguments"):
        # Because of the semantics of decorator composition if we just append
        # to the options list positional options will appear to be backwards.
        getattr(func, "arguments").insert(0, (args, kwargs))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Callable
import logging as log
from typing import Any
from urllib import parse as parser

from osprofiler import profiler
from osprofiler import web


# Register an OSProfiler HTTP Adapter that will profile any call made with
# requests.

LOG = log.getLogger(__name__)

_FUNC: Callable[..., Any] | None = None

try:
    from requests.adapters import HTTPAdapter
except ImportError:
    pass
else:

    def send(self: Any, request: Any, *args: Any, **kwargs: Any) -> Any:
        parsed_url = parser.urlparse(request.url)

        # Best effort guessing port if needed
        port: int | str = parsed_url.port or ""
        if not port and parsed_url.scheme == "http":
            port = 80
        elif not port and parsed_url.scheme == "https":
            port = 443

        profiler.start(
            parsed_url.scheme,
            info={
                "requests": {
                    "method": request.method,
                    "query": parsed_url.query,
                    "path": parsed_url.path,
                    "hostname": parsed_url.hostname,
                    "port": port,
                    "scheme": parsed_url.scheme,
                }
            },
        )

        # Profiling headers are overrident to take in account this new
        # context/span.
        request.headers.update(web.get_trace_id_headers())

        if _FUNC is None:
            raise RuntimeError("osprofiler requests adapter not initialized")
        response = _FUNC(self, request, *args, **kwargs)

        profiler.stop(info={"requests": {"status_code": response.status_code}})

        return response

    _FUNC = HTTPAdapter.send


def enable() -> None:
    if _FUNC:
        HTTPAdapter.send = send  # type: ignore[method-assign]
        LOG.debug("profiling requests enabled")
    else:
        LOG.warning(
            "unable to activate profiling for requests, "
            "please ensure that python requests is installed."
        )



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import datetime
import re
import unittest
from typing import Any, ClassVar
from unittest import mock


from osprofiler import profiler
from osprofiler.tests import test


class ProfilerGlobMethodsTestCase(test.TestCase):
    def test_get_profiler_not_inited(self):
        profiler.clean()
        self.assertIsNone(profiler.get())

    def test_get_profiler_and_init(self):
        p = profiler.init("secret", base_id="1", parent_id="2")
        self.assertEqual(profiler.get(), p)

        self.assertEqual(p.get_base_id(), "1")
        # NOTE(boris-42): until we make first start we don't have
        self.assertEqual(p.get_id(), "2")

    def test_start_not_inited(self):
        profiler.clean()
        profiler.start("name")

    def test_start(self):
        p = profiler.init("secret", base_id="1", parent_id="2")
        p.start = mock.MagicMock()  # type: ignore[method-assign]
        profiler.start("name", info="info")  # type: ignore[arg-type]
        p.start.assert_called_once_with("name", info="info")

    def test_stop_not_inited(self):
        profiler.clean()
        profiler.stop()

    def test_stop(self):
        p = profiler.init("secret", base_id="1", parent_id="2")
        p.stop = mock.MagicMock()  # type: ignore[method-assign]
        profiler.stop(info="info")  # type: ignore[arg-type]
        p.stop.assert_called_once_with(info="info")


class ProfilerTestCase(test.TestCase):
    def test_profiler_get_shorten_id(self):
        uuid_id = "4e3e0ec6-2938-40b1-8504-09eb1d4b0dee"
        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        result = prof.get_shorten_id(uuid_id)
        expected = "850409eb1d4b0dee"
        self.assertEqual(expected, result)

    def test_profiler_get_shorten_id_int(self):
        short_id_int = 42
        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        result = prof.get_shorten_id(short_id_int)
        expected = "2a"
        self.assertEqual(expected, result)

    def test_profiler_get_base_id(self):
        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        self.assertEqual(prof.get_base_id(), "1")

    @mock.patch("osprofiler.profiler.uuidutils.generate_uuid")
    def test_profiler_get_parent_id(self, mock_generate_uuid):
        mock_generate_uuid.return_value = "42"
        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        prof.start("test")
        self.assertEqual(prof.get_parent_id(), "2")

    @mock.patch("osprofiler.profiler.uuidutils.generate_uuid")
    def test_profiler_get_base_id_unset_case(self, mock_generate_uuid):
        mock_generate_uuid.return_value = "42"
        prof = profiler._Profiler("secret")
        self.assertEqual(prof.get_base_id(), "42")
        self.assertEqual(prof.get_parent_id(), "42")

    @mock.patch("osprofiler.profiler.uuidutils.generate_uuid")
    def test_profiler_get_id(self, mock_generate_uuid):
        mock_generate_uuid.return_value = "43"
        prof = profiler._Profiler("secret")
        prof.start("test")
        self.assertEqual(prof.get_id(), "43")

    @mock.patch("osprofiler.profiler.timeutils")
    @mock.patch("osprofiler.profiler.uuidutils.generate_uuid")
    @mock.patch("osprofiler.profiler.notifier.notify")
    def test_profiler_start(
        self, mock_notify, mock_generate_uuid, mock_timeutils
    ):
        mock_generate_uuid.return_value = "44"
        now = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        mock_timeutils.utcnow.return_value = now

        info = {"some": "info"}
        payload = {
            "name": "test-start",
            "base_id": "1",
            "parent_id": "2",
            "trace_id": "44",
            "info": info,
            "timestamp": now.strftime("%Y-%m-%dT%H:%M:%S.%f"),
        }

        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        prof.start("test", info=info)

        mock_notify.assert_called_once_with(payload)

    @mock.patch("osprofiler.profiler.timeutils")
    @mock.patch("osprofiler.profiler.notifier.notify")
    def test_profiler_stop(self, mock_notify, mock_timeutils):
        now = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        mock_timeutils.utcnow.return_value = now
        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        prof._trace_stack.append("44")
        prof._name.append("abc")

        info = {"some": "info"}
        prof.stop(info=info)

        payload = {
            "name": "abc-stop",
            "base_id": "1",
            "parent_id": "2",
            "trace_id": "44",
            "info": info,
            "timestamp": now.strftime("%Y-%m-%dT%H:%M:%S.%f"),
        }

        mock_notify.assert_called_once_with(payload)
        self.assertEqual(len(prof._name), 0)
        self.assertEqual(prof._trace_stack, collections.deque(["1", "2"]))

    @mock.patch("osprofiler.profiler.notifier.notify")
    def test_profiler_stop_empty_stack(self, mock_notify):
        """Test stop() gracefully handles empty profiler stack"""
        prof = profiler._Profiler("secret", base_id="1", parent_id="2")
        # Don't append anything to _name or _trace_stack
        # This simulates stop() being called without matching start()

        # Should not raise IndexError, just return silently
        prof.stop(info={"some": "info"})

        # No notification should be sent
        mock_notify.assert_not_called()
        # Stacks should remain empty
        self.assertEqual(len(prof._name), 0)
        self.assertEqual(len(prof._trace_stack), 2)  # base_id and parent_id

    def test_profiler_hmac(self):
        hmac = "secret"
        prof = profiler._Profiler(hmac, base_id="1", parent_id="2")
        self.assertEqual(hmac, prof.hmac_key)


class WithTraceTestCase(test.TestCase):
    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_with_trace(self, mock_start, mock_stop):

        with profiler.Trace("a", info="a1"):  # type: ignore[arg-type]
            mock_start.assert_called_once_with("a", info="a1")
            mock_start.reset_mock()
            with profiler.Trace("b", info="b1"):  # type: ignore[arg-type]
                mock_start.assert_called_once_with("b", info="b1")
            mock_stop.assert_called_once_with(info=None)
            mock_stop.reset_mock()
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_with_trace_etype(self, mock_start, mock_stop):

        def foo():
            with profiler.Trace("foo"):
                raise ValueError("bar")

        self.assertRaises(ValueError, foo)
        mock_start.assert_called_once_with("foo", info=None)
        mock_stop.assert_called_once_with(
            info={"etype": "ValueError", "message": "bar"}
        )


@profiler.trace("function", info={"info": "some_info"})
def traced_func(i):
    return i


@profiler.trace("hide_args", hide_args=True)
def trace_hide_args_func(a, i=10):
    return (a, i)


@profiler.trace("foo", hide_args=True)
def test_fn_exc():
    raise ValueError()


@profiler.trace("hide_result", hide_result=False)
def trace_with_result_func(a, i=10):
    return (a, i)


class TraceDecoratorTestCase(test.TestCase):
    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_duplicate_trace_disallow(self, mock_start, mock_stop):

        @profiler.trace("test")
        def trace_me():
            pass

        self.assertRaises(
            ValueError,
            profiler.trace("test-again", allow_multiple_trace=False),
            trace_me,
        )

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_with_args(self, mock_start, mock_stop):
        self.assertEqual(1, traced_func(1))
        expected_info = {
            "info": "some_info",
            "function": {
                "name": "osprofiler.tests.unit.test_profiler.traced_func",
                "args": str((1,)),
                "kwargs": str({}),
            },
        }
        mock_start.assert_called_once_with("function", info=expected_info)
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_without_args(self, mock_start, mock_stop):
        self.assertEqual((1, 2), trace_hide_args_func(1, i=2))
        expected_info = {
            "function": {
                "name": "osprofiler.tests.unit.test_profiler"
                ".trace_hide_args_func"
            }
        }
        mock_start.assert_called_once_with("hide_args", info=expected_info)
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_with_exception(self, mock_start, mock_stop):

        self.assertRaises(ValueError, test_fn_exc)
        expected_info = {
            "function": {
                "name": "osprofiler.tests.unit.test_profiler.test_fn_exc"
            }
        }
        expected_stop_info = {"etype": "ValueError", "message": ""}
        mock_start.assert_called_once_with("foo", info=expected_info)
        mock_stop.assert_called_once_with(info=expected_stop_info)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_with_result(self, mock_start, mock_stop):
        self.assertEqual((1, 2), trace_with_result_func(1, i=2))
        start_info = {
            "function": {
                "name": "osprofiler.tests.unit.test_profiler"
                ".trace_with_result_func",
                "args": str((1,)),
                "kwargs": str({"i": 2}),
            }
        }

        stop_info = {"function": {"result": str((1, 2))}}
        mock_start.assert_called_once_with("hide_result", info=start_info)
        mock_stop.assert_called_once_with(info=stop_info)


class FakeTracedCls:
    def method1(self, a, b, c=10):
        return a + b + c

    def method2(self, d, e):
        return d - e

    def method3(self, g=10, h=20):
        return g * h

    def _method(self, i):
        return i


@profiler.trace_cls("rpc", info={"a": 10})
class FakeTraceClassWithInfo(FakeTracedCls):
    pass


@profiler.trace_cls("a", info={"b": 20}, hide_args=True)
class FakeTraceClassHideArgs(FakeTracedCls):
    pass


@profiler.trace_cls("rpc", trace_private=True)
class FakeTracePrivate(FakeTracedCls):
    pass


class FakeTraceStaticMethodBase(FakeTracedCls):
    @staticmethod
    def static_method(arg):
        return arg


@profiler.trace_cls("rpc", trace_static_methods=True)
class FakeTraceStaticMethod(FakeTraceStaticMethodBase):
    pass


@profiler.trace_cls("rpc")
class FakeTraceStaticMethodSkip(FakeTraceStaticMethodBase):
    pass


class FakeTraceClassMethodBase(FakeTracedCls):
    @classmethod
    def class_method(cls, arg):
        return arg


@profiler.trace_cls("rpc")
class FakeTraceClassMethodSkip(FakeTraceClassMethodBase):
    pass


def py3_info(info):
    # NOTE(boris-42): py33 I hate you.
    info_py3 = copy.deepcopy(info)
    new_name = re.sub(
        "FakeTrace[^.]*", "FakeTracedCls", info_py3["function"]["name"]
    )
    info_py3["function"]["name"] = new_name
    return info_py3


def possible_mock_calls(name, info):
    # NOTE(boris-42): py33 I hate you.
    return [mock.call(name, info=info), mock.call(name, info=py3_info(info))]


class TraceClsDecoratorTestCase(test.TestCase):
    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_args(self, mock_start, mock_stop):
        fake_cls = FakeTraceClassWithInfo()
        self.assertEqual(30, fake_cls.method1(5, 15))
        expected_info = {
            "a": 10,
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceClassWithInfo.method1"
                ),
                "args": str((fake_cls, 5, 15)),
                "kwargs": str({}),
            },
        }
        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_kwargs(self, mock_start, mock_stop):
        fake_cls = FakeTraceClassWithInfo()
        self.assertEqual(50, fake_cls.method3(g=5, h=10))
        expected_info = {
            "a": 10,
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceClassWithInfo.method3"
                ),
                "args": str((fake_cls,)),
                "kwargs": str({"g": 5, "h": 10}),
            },
        }
        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_without_private(self, mock_start, mock_stop):
        fake_cls = FakeTraceClassHideArgs()
        self.assertEqual(10, fake_cls._method(10))
        mock_start.assert_not_called()
        mock_stop.assert_not_called()

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_without_args(self, mock_start, mock_stop):
        fake_cls = FakeTraceClassHideArgs()
        self.assertEqual(40, fake_cls.method1(5, 15, c=20))
        expected_info = {
            "b": 20,
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceClassHideArgs.method1"
                ),
            },
        }

        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("a", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_private_methods(self, mock_start, mock_stop):
        fake_cls = FakeTracePrivate()
        self.assertEqual(5, fake_cls._method(5))

        expected_info = {
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTracePrivate._method"
                ),
                "args": str((fake_cls, 5)),
                "kwargs": str({}),
            }
        }

        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    @unittest.skip(
        "Static method tracing was disabled due the bug. This test should be "
        "skipped until we find the way to address it."
    )
    def test_static(self, mock_start, mock_stop):
        fake_cls = FakeTraceStaticMethod()

        self.assertEqual(25, fake_cls.static_method(25))

        expected_info = {
            "function": {
                # fixme(boris-42): Static methods are treated differently in
                #                  Python 2.x and Python 3.x. So in PY2 we
                #                  expect to see method4 because method is
                #                  static and doesn't have reference to class
                #                  - and FakeTraceStatic.method4 in PY3
                "name": "osprofiler.tests.unit.test_profiler"
                "osprofiler.tests.unit.test_profiler.FakeTraceStatic"
                ".method4",
                "args": str((25,)),
                "kwargs": str({}),
            }
        }

        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_static_method_skip(self, mock_start, mock_stop):
        self.assertEqual(25, FakeTraceStaticMethodSkip.static_method(25))
        mock_start.assert_not_called()
        mock_stop.assert_not_called()

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_class_method_skip(self, mock_start, mock_stop):
        self.assertEqual("foo", FakeTraceClassMethodSkip.class_method("foo"))
        mock_start.assert_not_called()
        mock_stop.assert_not_called()


class FakeTraceWithMetaclassBase(metaclass=profiler.TracedMeta):
    __trace_args__: ClassVar[dict[str, Any]] = {
        "name": "rpc",
        "info": {"a": 10},
    }

    def method1(self, a, b, c=10):
        return a + b + c

    def method2(self, d, e):
        return d - e

    def method3(self, g=10, h=20):
        return g * h

    def _method(self, i):
        return i


class FakeTraceDummy(FakeTraceWithMetaclassBase):
    def method4(self, j):
        return j


class FakeTraceWithMetaclassHideArgs(FakeTraceWithMetaclassBase):
    __trace_args__: ClassVar[dict[str, Any]] = {
        "name": "a",
        "info": {"b": 20},
        "hide_args": True,
    }

    def method5(self, k, l):
        return k + l


class FakeTraceWithMetaclassPrivate(FakeTraceWithMetaclassBase):
    __trace_args__: ClassVar[dict[str, Any]] = {
        "name": "rpc",
        "trace_private": True,
    }

    def _new_private_method(self, m):
        return 2 * m


class TraceWithMetaclassTestCase(test.TestCase):
    def test_no_name_exception(self):
        def define_class_with_no_name():
            class FakeTraceWithMetaclassNoName(
                FakeTracedCls, metaclass=profiler.TracedMeta
            ):
                pass

        self.assertRaises(TypeError, define_class_with_no_name, 1)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_args(self, mock_start, mock_stop):
        fake_cls = FakeTraceWithMetaclassBase()
        self.assertEqual(30, fake_cls.method1(5, 15))
        expected_info = {
            "a": 10,
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceWithMetaclassBase.method1"
                ),
                "args": str((fake_cls, 5, 15)),
                "kwargs": str({}),
            },
        }
        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_kwargs(self, mock_start, mock_stop):
        fake_cls = FakeTraceWithMetaclassBase()
        self.assertEqual(50, fake_cls.method3(g=5, h=10))
        expected_info = {
            "a": 10,
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceWithMetaclassBase.method3"
                ),
                "args": str((fake_cls,)),
                "kwargs": str({"g": 5, "h": 10}),
            },
        }
        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_without_private(self, mock_start, mock_stop):
        fake_cls = FakeTraceWithMetaclassHideArgs()
        self.assertEqual(10, fake_cls._method(10))
        mock_start.assert_not_called()
        mock_stop.assert_not_called()

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_without_args(self, mock_start, mock_stop):
        fake_cls = FakeTraceWithMetaclassHideArgs()
        self.assertEqual(20, fake_cls.method5(5, 15))
        expected_info = {
            "b": 20,
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceWithMetaclassHideArgs.method5"
                )
            },
        }

        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("a", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)

    @mock.patch("osprofiler.profiler.stop")
    @mock.patch("osprofiler.profiler.start")
    def test_private_methods(self, mock_start, mock_stop):
        fake_cls = FakeTraceWithMetaclassPrivate()
        self.assertEqual(10, fake_cls._new_private_method(5))

        expected_info = {
            "function": {
                "name": (
                    "osprofiler.tests.unit.test_profiler"
                    ".FakeTraceWithMetaclassPrivate._new_private_method"
                ),
                "args": str((fake_cls, 5)),
                "kwargs": str({}),
            }
        }

        mock_start.assert_called_once()
        self.assertIn(
            mock_start.call_args_list[0],
            possible_mock_calls("rpc", expected_info),
        )
        mock_stop.assert_called_once_with(info=None)



# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from osprofiler.drivers.elasticsearch_driver import ElasticsearchDriver
from osprofiler.tests import test


class ElasticsearchTestCase(test.TestCase):
    def setUp(self):
        super().setUp()
        self.elasticsearch = ElasticsearchDriver(
            "elasticsearch://localhost:9001/"
        )
        self.elasticsearch.project = "project"
        self.elasticsearch.service = "service"

    def test_init_and_notify(self):
        self.elasticsearch.client = mock.MagicMock()
        self.elasticsearch.client.reset_mock()
        project = "project"
        service = "service"
        host = "host"

        info = {"a": 10, "project": project, "service": service, "host": host}
        self.elasticsearch.notify(info)

        self.elasticsearch.client.index.assert_called_once_with(
            index="osprofiler-notifications",
            doc_type="notification",
            body=info,
        )

    def test_get_empty_report(self):
        self.elasticsearch.client = mock.MagicMock()
        self.elasticsearch.client.search = mock.MagicMock(
            return_value={"_scroll_id": "1", "hits": {"hits": []}}
        )
        self.elasticsearch.client.reset_mock()

        get_report = self.elasticsearch.get_report
        base_id = "abacaba"

        get_report(base_id)

        self.elasticsearch.client.search.assert_called_once_with(
            index="osprofiler-notifications",
            doc_type="notification",
            size=10000,
            scroll="2m",
            body={"query": {"match": {"base_id": base_id}}},
        )

    def test_get_non_empty_report(self):
        base_id = "1"
        elasticsearch_first_response = {
            "_scroll_id": "1",
            "hits": {
                "hits": [
                    {
                        "_source": {
                            "timestamp": "2016-08-10T16:58:03.064438",
                            "base_id": base_id,
                            "project": "project",
                            "service": "service",
                            "parent_id": "0",
                            "name": "test",
                            "info": {"host": "host"},
                            "trace_id": "1",
                        }
                    }
                ]
            },
        }
        elasticsearch_second_response = {
            "_scroll_id": base_id,
            "hits": {"hits": []},
        }
        self.elasticsearch.client = mock.MagicMock()
        self.elasticsearch.client.search = mock.MagicMock(
            return_value=elasticsearch_first_response
        )
        self.elasticsearch.client.scroll = mock.MagicMock(
            return_value=elasticsearch_second_response
        )

        self.elasticsearch.client.reset_mock()

        self.elasticsearch.get_report(base_id)

        self.elasticsearch.client.search.assert_called_once_with(
            index="osprofiler-notifications",
            doc_type="notification",
            size=10000,
            scroll="2m",
            body={"query": {"match": {"base_id": base_id}}},
        )

        self.elasticsearch.client.scroll.assert_called_once_with(
            scroll_id=base_id, scroll="2m"
        )



# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import fixture

from osprofiler import opts
from osprofiler.tests import test


class ConfigTestCase(test.TestCase):
    def setUp(self):
        super().setUp()
        self.conf_fixture = self.useFixture(fixture.Config())

    def test_options_defaults(self):
        opts.set_defaults(self.conf_fixture.conf)
        self.assertFalse(self.conf_fixture.conf.profiler.enabled)
        self.assertFalse(self.conf_fixture.conf.profiler.trace_sqlalchemy)
        self.assertEqual(
            "SECRET_KEY", self.conf_fixture.conf.profiler.hmac_keys
        )
        self.assertFalse(opts.is_trace_enabled(self.conf_fixture.conf))
        self.assertFalse(opts.is_db_trace_enabled(self.conf_fixture.conf))

    def test_options_defaults_override(self):
        opts.set_defaults(
            self.conf_fixture.conf,
            enabled=True,
            trace_sqlalchemy=True,
            hmac_keys="MY_KEY",
        )
        self.assertTrue(self.conf_fixture.conf.profiler.enabled)
        self.assertTrue(self.conf_fixture.conf.profiler.trace_sqlalchemy)
        self.assertEqual("MY_KEY", self.conf_fixture.conf.profiler.hmac_keys)
        self.assertTrue(opts.is_trace_enabled(self.conf_fixture.conf))
        self.assertTrue(opts.is_db_trace_enabled(self.conf_fixture.conf))

    @mock.patch("osprofiler.web.enable")
    @mock.patch("osprofiler.web.disable")
    def test_web_trace_disabled(self, mock_disable, mock_enable):
        opts.set_defaults(self.conf_fixture.conf, hmac_keys="MY_KEY")
        opts.enable_web_trace(self.conf_fixture.conf)
        opts.disable_web_trace(self.conf_fixture.conf)
        mock_enable.assert_not_called()
        mock_disable.assert_not_called()

    @mock.patch("osprofiler.web.enable")
    @mock.patch("osprofiler.web.disable")
    def test_web_trace_enabled(self, mock_disable, mock_enable):
        opts.set_defaults(
            self.conf_fixture.conf, enabled=True, hmac_keys="MY_KEY"
        )
        opts.enable_web_trace(self.conf_fixture.conf)
        opts.disable_web_trace(self.conf_fixture.conf)
        mock_enable.assert_called_once_with("MY_KEY")
        mock_disable.assert_called_once_with()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from osprofiler import _utils as utils
from osprofiler.drivers import otlp
from osprofiler import opts
from osprofiler.tests import test


class OTLPTestCase(test.TestCase):
    def setUp(self):
        super().setUp()

        opts.set_defaults(cfg.CONF)

        self.payload_start = {
            "name": "api-start",
            "base_id": "4e3e0ec6-2938-40b1-8504-09eb1d4b0dee",
            "trace_id": "1c089ea8-28fe-4f3d-8c00-f6daa2bc32f1",
            "parent_id": "e2715537-3d1c-4f0c-b3af-87355dc5fc5b",
            "timestamp": "2018-05-03T04:31:51.781381",
            "info": {"host": "test"},
        }

        self.root_span_payload_start = {
            "name": "api-start",
            "base_id": "4e3e0ec6-2938-40b1-8504-09eb1d4b0dee",
            "trace_id": "1c089ea8-28fe-4f3d-8c00-f6daa2bc32f1",
            "parent_id": "4e3e0ec6-2938-40b1-8504-09eb1d4b0dee",
            "timestamp": "2018-05-03T04:31:51.781381",
            "info": {"host": "test"},
        }

        self.payload_stop = {
            "name": "api-stop",
            "base_id": "4e3e0ec6-2938-40b1-8504-09eb1d4b0dee",
            "trace_id": "1c089ea8-28fe-4f3d-8c00-f6daa2bc32f1",
            "parent_id": "e2715537-3d1c-4f0c-b3af-87355dc5fc5b",
            "timestamp": "2018-05-03T04:31:51.781381",
            "info": {"host": "test", "function": {"result": 1}},
        }

        self.driver = otlp.OTLP(
            "otlp://127.0.0.1:6831",
            project="nova",
            service="api",
            conf=cfg.CONF,
        )

    def test_notify_start(self):
        self.driver.notify(self.payload_start)
        self.assertEqual(1, len(self.driver.spans))

    def test_notify_start_with_parent_span(self):
        self.driver.notify(self.payload_start)
        self.assertEqual(1, len(self.driver.spans))
        result_span = self.driver.spans[0]
        self.assertEqual(
            utils.uuid_to_int128(str(self.payload_start["base_id"])),
            result_span._parent.trace_id,
        )
        self.assertEqual(
            utils.shorten_id(str(self.payload_start["parent_id"])),
            result_span._parent.span_id,
        )

    def test_notify_start_with_root_span(self):
        self.driver.notify(self.root_span_payload_start)
        self.assertEqual(1, len(self.driver.spans))
        result_span = self.driver.spans[0]
        self.assertIsNone(result_span._parent)

    def test_notify_stop(self):
        mock_end = mock.MagicMock()
        self.driver.notify(self.payload_start)
        self.driver.spans[0].end = mock_end
        self.driver.notify(self.payload_stop)
        mock_end.assert_called_once()

    def test_notify_stop_with_db_result(self):
        self.payload_stop["info"] = {"host": "test", "db": {"result": "()"}}
        mock_end = mock.MagicMock()
        self.driver.notify(self.payload_start)
        self.driver.spans[0].end = mock_end
        self.driver.notify(self.payload_stop)
        mock_end.assert_called_once()

    def test_notify_stop_with_db_error(self):
        self.payload_stop["info"] = {
            "host": "test",
            "etype": "test_error",
            "message": "test_error_message",
            "db": {
                "original_exception": "test_exception",
                "chained_exception": "test_chained_exception",
            },
        }
        mock_end = mock.MagicMock()
        self.driver.notify(self.payload_start)
        self.driver.spans[0].end = mock_end
        self.driver.notify(self.payload_stop)
        mock_end.assert_called_once()

    def test_notify_stop_without_function_result(self):
        self.payload_stop["info"] = {"host": "test", "function": {}}
        mock_end = mock.MagicMock()
        self.driver.notify(self.payload_start)
        self.driver.spans[0].end = mock_end
        self.driver.notify(self.payload_stop)
        mock_end.assert_called_once()

    def test_notify_stop_with_requests(self):
        self.payload_stop["info"] = {
            "host": "test",
            "requests": {
                "status_code": 200,
            },
        }
        mock_end = mock.MagicMock()
        self.driver.notify(self.payload_start)
        self.driver.spans[0].end = mock_end
        self.driver.notify(self.payload_stop)
        mock_end.assert_called_once()

    def test_service_name_default(self):
        self.assertEqual(
            "pr1-svc1", self.driver._get_service_name(cfg.CONF, "pr1", "svc1")
        )

    def test_service_name_prefix(self):
        cfg.CONF.set_default("service_name_prefix", "prx1", "profiler_otlp")
        self.assertEqual(
            "prx1-pr1-svc1",
            self.driver._get_service_name(cfg.CONF, "pr1", "svc1"),
        )

    def test_process_tags(self):
        # Need to be implemented.
        pass



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Generator
import contextlib
import logging as log
from typing import Any

from oslo_utils import reflection

from osprofiler import profiler

LOG = log.getLogger(__name__)

_DISABLED = False


def disable() -> None:
    """Disable tracing of all DB queries. Reduce a lot size of profiles."""
    global _DISABLED
    _DISABLED = True


def enable() -> None:
    """add_tracing adds event listeners for sqlalchemy."""

    global _DISABLED
    _DISABLED = False


def add_tracing(
    sqlalchemy: Any, engine: Any, name: str, hide_result: bool = True
) -> None:
    """Add tracing to all sqlalchemy calls."""

    if not _DISABLED:
        sqlalchemy.event.listen(
            engine, "before_cursor_execute", _before_cursor_execute(name)
        )
        sqlalchemy.event.listen(
            engine,
            "after_cursor_execute",
            _after_cursor_execute(hide_result=hide_result),
        )
        sqlalchemy.event.listen(engine, "handle_error", handle_error)


@contextlib.contextmanager
def wrap_session(sqlalchemy: Any, sess: Any) -> Generator[Any, None, None]:
    with sess as s:
        if not getattr(s.bind, "traced", False):
            add_tracing(sqlalchemy, s.bind, "db")
            s.bind.traced = True
        yield s


def _before_cursor_execute(name: str) -> Any:
    """Add listener that will send trace info before query is executed."""

    def handler(
        conn: Any,
        cursor: Any,
        statement: Any,
        params: Any,
        context: Any,
        executemany: Any,
    ) -> None:
        info = {"db": {"statement": statement, "params": params}}
        profiler.start(name, info=info)

    return handler


def _after_cursor_execute(hide_result: bool = True) -> Any:
    """Add listener that will send trace info after query is executed.

    :param hide_result: Boolean value to hide or show SQL result in trace.
                        True - hide SQL result (default).
                        False - show SQL result in trace.
    """

    def handler(
        conn: Any,
        cursor: Any,
        statement: Any,
        params: Any,
        context: Any,
        executemany: Any,
    ) -> None:
        if not hide_result:
            # Add SQL result to trace info in *-stop phase
            info = {"db": {"result": str(cursor._rows)}}
            profiler.stop(info=info)
        else:
            profiler.stop()

    return handler


def handle_error(exception_context: Any) -> None:
    """Handle SQLAlchemy errors"""
    exception_class_name = reflection.get_class_name(
        exception_context.original_exception
    )
    original_exception = str(exception_context.original_exception)

    cause = exception_context.original_exception.__cause__
    chained_exception = str(cause) if cause else ''

    info = {
        "etype": exception_class_name,
        "message": original_exception,
        "db": {
            "original_exception": original_exception,
            "chained_exception": chained_exception,
        },
    }
    profiler.stop(info=info)
    LOG.debug(
        "OSProfiler has handled SQLAlchemy error: %s", original_exception
    )
